<?php
// This file is part of Stack - http://stack.maths.ed.ac.uk/
//
// Stack is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Stack is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Stack.  If not, see <http://www.gnu.org/licenses/>.

/**
 * This script runs the answers tests and verifies the results.
 *
 * This serves two purposes. First, it verifies that the answer tests are working
 * correctly, and second it serves to document the expected behaviour of answer
 * tests, which is useful for learning how they work.
 *
 * @copyright  2012 University of Birmingham
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

require_once('equivfixtures.class.php');

class stack_answertest_test_data {
    const NAME    = 0;
    const OPTIONS = 1;
    const SANS    = 2;
    const TANS    = 3;
    const SCORE   = 4;
    const ANSNOTE = 5;
    const NOTES   = 6;

    /**
     * Raw data should be in the following form (matching the constants above).
     * Name of test
     * Options
     * Student's answer
     * Teacher's answer
     * Expected result
     *    0 = different
     *    1 = same
     *    -1 = "Test fails", but this is trapped.
     *    -2 = Expected maths failure, where test should return 0.
     *    -3 = Expected maths failure, where test should return 1.
     * Answer note(s)
     * Comments on this test.
     * Header row in the table (optional).
     */
    protected static $rawdata = array(

        array('AlgEquiv', '', '1/0', '1', -1, 'ATAlgEquiv_STACKERROR_SAns.', ''),
        array('AlgEquiv', '', '1', '1/0', -1, 'ATAlgEquiv_STACKERROR_TAns.', ''),
        array('AlgEquiv', '', '', '(x-1)^2', -1, 'ATAlgEquivTEST_FAILED-Empty SA.', ''),
        array('AlgEquiv', '', 'x^2', '', -1, 'ATAlgEquivTEST_FAILED-Empty TA.', ''),
        array('AlgEquiv', '', 'x-1)^2', '(x-1)^2', -1, 'ATAlgEquivTEST_FAILED-Empty SA.', ''),

        // Make this behaviour explict.
        array('AlgEquiv', '', 'x1', 'x_1', 0, '', ''),
        array('AlgEquiv', '', 'x_1', 'x[1]', 0, '', ''),
        array('AlgEquiv', '', 'x[1]', 'x1', 0, '', ''),
        array('AlgEquiv', '', 'integerp(3)', 'true', 1, 'ATLogic_True.', 'Predicates'),
        array('AlgEquiv', '', 'integerp(3.1)', 'true', 0, '', ''),
        array('AlgEquiv', '', 'integerp(3)', 'false', 0, '', ''),
        array('AlgEquiv', '', 'integerp(3)', 'true', 1, 'ATLogic_True.', ''),
        // Note that AlgEquiv simplifies its arguments.  Use a non-simplifying test instead.
        array('AlgEquiv', '', 'lowesttermsp(x^2/x)', 'true', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'lowesttermsp(-y/-x)', 'true', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'lowesttermsp((x^2-1)/(x-1))', 'true', 0, '', ''),
        array('AlgEquiv', '', 'lowesttermsp((x^2-1)/(x+2))', 'true', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'X', 'x', 0, 'ATAlgEquiv_WrongCase.', 'Case sensitivity'),
        array('AlgEquiv', '', 'exdowncase(X)', 'x', 1, '', ''),
        array('AlgEquiv', '', 'exdowncase((X-1)^2)', 'x^2-2*x+1', 1, '', ''),
        array('AlgEquiv', '', 'Y=1+X', 'y=1+x', 0, 'ATEquation_default',
            'Permutations of variables (To do: a dedicated answer test with feedback)'),
        array('AlgEquiv', '', 'v+w+x+y+z', 'a+b+c+A+B', 0, '', ''),
        array('AlgEquiv', '', '4^(-1/2)', '1/2', 1, '', 'Numbers'),
        array('AlgEquiv', '', '4^(1/2)', 'sqrt(4)', 1, '', ''),
        array('AlgEquiv', '', '0.5', '1/2', 1, '', 'Mix of floats and rational numbers'),
        array('AlgEquiv', '', '0.33', '1/3', 0, '', ''),
        array('AlgEquiv', '', '5.1e-2', '51/1000', 1, '', ''),
        array('AlgEquiv', '', '0.333333333333333', '1/3', 0, '', ''),
        array('AlgEquiv', '', '(0.5+x)*2', '2*x+1', 1, '', ''),

        array('AlgEquiv', '', 'sqrt(-1)', '%i', 1, '', 'Complex numbers'),
        array('AlgEquiv', '', '%i', 'e^(i*pi/2)', 1, '', ''),
        array('AlgEquiv', '', '(4*sqrt(3)*%i+4)^(1/5)', '8^(1/5)*(cos(%pi/15)+%i*sin(%pi/15))', 1, '', ''),
        array('AlgEquiv', '', '(4*sqrt(3)*%i+4)^(1/5)', 'rectform((4*sqrt(3)*%i+4)^(1/5))', 1, '', ''),
        array('AlgEquiv', '', '(4*sqrt(3)*%i+4)^(1/5)', 'polarform((4*sqrt(3)*%i+4)^(1/5))', 1, '', ''),
        array('AlgEquiv', '', '%i/sqrt(x)', 'sqrt(-1/x)', 1, '', ''),

        array('AlgEquiv', '', 'inf', 'inf', 1, '', 'Infinity'),
        array('AlgEquiv', '', 'inf', '-inf', 0, '', ''),
        array('AlgEquiv', '', '2*inf', 'inf', 0, '', ''),
        array('AlgEquiv', '', '0*inf', '0', 1, '', ''),

        array('AlgEquiv', '', 'x^(1/2)', 'sqrt(x)', 1, '', 'Powers and roots'),
        array('AlgEquiv', '', 'x', 'sqrt(x^2)', 0, '', ''),
        array('AlgEquiv', '', 'abs(x)', 'sqrt(x^2)', 1, '', ''),
        array('AlgEquiv', '', '1/abs(x)^(1/3)', '(abs(x)^(1/3)/abs(x))^(1/2)', 1, '', ''),
        array('AlgEquiv', '', 'sqrt((x-3)*(x-5))', 'sqrt(x-3)*sqrt(x-5)', 0, '', ''),
        array('AlgEquiv', '', '1/sqrt(x)', 'sqrt(1/x)', 1, '', ''),
        array('AlgEquiv', '', 'x-1', '(x^2-1)/(x+1)', 1, '', ''),
        array('AlgEquiv', '', '2^((1/5.1)*t)', '2^((1/5.1)*t)', 1, '', ''),
        array('AlgEquiv', '', '2^((1/5.1)*t)', '2^(0.196078431373*t)', 0, '', ''),
        array('AlgEquiv', '', 'a^b*a^c', 'a^(b+c)', 1, '', ''),
        array('AlgEquiv', '', '(assume(x>2),6*((x-2)^2)^k)', '6*(x-2)^(2*k)', 1, '', ''),
        array('AlgEquiv', '', '6*((x-2)^3)^k', '6*(x-2)^(3*k)', 1, '', ''),
        array('AlgEquiv', '', '(4*sqrt(3)*%i+4)^(1/5)', '6^(1/5)*cos(%pi/15)-6^(1/5)*%i*sin(%pi/15)', 0, '', ''),
        array('AlgEquiv', '', '2+2*sqrt(3+x)', '2+sqrt(12+4*x)', 1, '', ''),
        array('AlgEquiv', '', '6*e^(6*(y^2+x^2))+72*x^2*e^(6*(y^2+x^2))',
            '(72*x^2+6)*e^(6*(y^2+x^2))', 1, '', '', ''),
        array('AlgEquiv', '', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 1, '',
            'Expressions with subscripts'),
        array('AlgEquiv', '', 'rho*z*V/(4*pi*epsilon[1]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 0, '', ''),
        array('AlgEquiv', '', 'sqrt(k/m)*sqrt(m/k)', '1', 1, '', ''),
        array('AlgEquiv', '', '(2*pi)/(k/m)^(1/2)', '(2*pi)/(k/m)^(1/2)', 1, '', ''),
        array('AlgEquiv', '', '(2*pi)*(m/k)^(1/2)', '(2*pi)/(k/m)^(1/2)', 1, '', ''),
        array('AlgEquiv', '', 'sqrt(2*x/10+1)', 'sqrt((2*x+10)/10)', 1, '', ''),
        array('AlgEquiv', '', '((x+3)^2*(x+3))^(1/3)', '((x+3)*(x^2+6*x+9))^(1/3)', 1, '', ''),
        array('AlgEquiv', '', '((x+3)^2*(x+3))^(1/3)', '((x+3)*(x^2+6*x+9))^(1/3)', 1, '', 'Need to factor internally.'),

        array('AlgEquiv', '', '(x-1)^2', 'x^2-2*x+1', 1, '', 'Polynomials and rational function'),
        array('AlgEquiv', '', '(x-1)*(x^2+x+1)', 'x^3-1', 1, '', ''),
        array('AlgEquiv', '', '(x-1)^(-2)', '1/(x^2-2*x+1)', 1, '', ''),
        array('AlgEquiv', '', '1/(4*x-(%pi+sqrt(2)))', '1/(x+1)', 0, '', ''),
        array('AlgEquiv', '', '(x-a)^6000', '(x-a)^6000', 1, '', ''),
        array('AlgEquiv', '', '(a-x)^6000', '(x-a)^6000', 1, '', ''),
        array('AlgEquiv', '', '(4*a-x)^6000', '(x-4*a)^6000', 1, '', ''),
        array('AlgEquiv', '', '(x-a)^6000', '(x-a)^5999', 0, '', ''),
        array('AlgEquiv', '', '(k+8)/(k^2+4*k-12)', '(k+8)/(k^2+4*k-12)', 1, '', ''),
        array('AlgEquiv', '', '(k+7)/(k^2+4*k-12)', '(k+8)/(k^2+4*k-12)', 0, '', ''),
        array('AlgEquiv', '', '-(2*k+6)/(k^2+4*k-12)', '-(2*k+6)/(k^2+4*k-12)', 1, '', ''),
        array('AlgEquiv', '', '1/n-1/(n+1)', '1/(n*(n+1))', 1, '', ''),
        array('AlgEquiv', '', '0.5*x^2+3*x-1', 'x^2/2+3*x-1', 1, '', ''),

        array('AlgEquiv', '', 'cos(x)', 'cos(-x)', 1, '', 'Trig functions'),
        array('AlgEquiv', '', 'cos(x)^2+sin(x)^2', '1', 1, '', ''),
        array('AlgEquiv', '', 'cos(x+y)', 'cos(x)*cos(y)-sin(x)*sin(y)', 1, '', ''),
        array('AlgEquiv', '', 'cos(x+y)', 'cos(x)*cos(y)+sin(x)*sin(y)', 0, '', ''),
        array('AlgEquiv', '', 'cos(x#pm#y)', 'cos(x)*cos(y)-(#pm#sin(x)*sin(y))', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'sin(x#pm#y)', 'sin(x)*cos(y)#pm#cos(x)*sin(y)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'sin(x#pm#y)', 'cos(x)*sin(y)#pm#sin(x)*cos(y)', 0, '', ''),
        array('AlgEquiv', '', '2*cos(x)^2-1', 'cos(2*x)', 1, '', ''),
        array('AlgEquiv', '', '1.0*cos(1200*%pi*x)', 'cos(1200*%pi*x)', 1, '', ''),
        array('AlgEquiv', '', 'diff(tan(10*x)^2,x)', 'cos(6*x)', 0, '', ''),
        array('AlgEquiv', '', 'exp(%i*%pi)', '-1', 1, '', ''),
        array('AlgEquiv', '', '2*cos(2*x)+x+1', '-sin(x)^2+3*cos(x)^2+x', 1, '', ''),

        array('AlgEquiv', '', '(2*sec(2*t)^2-2)/2',
            '-(sin(4*t)^2-2*sin(4*t)+cos(4*t)^2-1)*(sin(4*t)^2+2*sin(4*t)+cos(4*t)^2-1)/(sin(4*t)^2+cos(4*t)^2+2*cos(4*t)+1)^2',
            1, '', ''),
        array('AlgEquiv', '', '1+cosec(3*x)', '1+csc(3*x)', 1, '', ''),
        array('AlgEquiv', '', '1/(1+exp(-2*x))', 'tanh(x)/2+1/2', 1, '', ''),
        array('AlgEquiv', '', '1+cosech(3*x)', '1+csch(3*x)', 1, '', ''),
        array('AlgEquiv', '', '-4*sec(4*z)^2*sin(6*z)-6*tan(4*z)*cos(6*z)',
            '-4*sec(4*z)^2*sin(6*z)-6*tan(4*z)*cos(6*z)', 1, '', ''),
        array('AlgEquiv', '', '-4*sec(4*z)^2*sin(6*z)-6*tan(4*z)*cos(6*z)',
            '4*sec(4*z)^2*sin(6*z)+6*tan(4*z)*cos(6*z)', 0, '', ''),
        // The following test is here because we can't factor with trigsimp:true.
        array('AlgEquiv', '', 'csc(6*x)^2*(7*sin(6*x)*cos(7*x)-6*cos(6*x)*sin(7*x))',
            '-(6*cos(6*x)*sin(7*x)-7*sin(6*x)*cos(7*x))/sin(6*x)^2', 1, '', ''),
        array('AlgEquiv', '', 'csc(6*x)^2*(7*sin(6*x)*cos(7*x)-6*cos(6*x)*sin(7*x))',
            '(6*cos(6*x)*sin(7*x)-7*sin(6*x)*cos(7*x))/sin(6*x)^2', 0, '', ''),
        array('AlgEquiv', '', '-(7*x^6+4*x^3)/sin(7*y+x^7+x^4+1)^2',
            '-(7*x^6+4*x^3)*csc(7*y+x^7+x^4+1)^2', 1, '', ''),
        array('AlgEquiv', '', 'sin((2*%pi*n-%pi)/2)', '-cos(n*%pi)', 1, '', ''),
        array('AlgEquiv', '', 'sin(x/2)/(1+tan(x)*tan(x/2))', 'sin(x/2)*cos(x)', 1, '', ''),
        array('AlgEquiv', '', '(declare(n,integer),trigrat(sin((2*%pi*n-%pi)/2)))', '-(-1)^n', 1, '', ''),
        // According to Twitter!  Not sure this is even true, but just for fun!
        array('AlgEquiv', '', 'cot(%pi/20)+cot(%pi/24)-cot(%pi/10)', 'sqrt(1)+sqrt(2)+sqrt(3)+sqrt(4)+sqrt(5)+sqrt(6)',
            -3, '', '', ''),
        array('AlgEquiv', '', 'trigeval(cot(%pi/20)+cot(%pi/24)-cot(%pi/10))', 'sqrt(1)+sqrt(2)+sqrt(3)+sqrt(4)+sqrt(5)+sqrt(6)',
            1, '', '', ''),
        array('AlgEquiv', '', 'sin([1/8,1/6, 1/4, 1/3, 1/2, 1]*%pi)', '[sqrt(2-sqrt(2))/2,1/2,1/sqrt(2),sqrt(3)/2,1,0]',
            -3, '(ATList_wrongentries 1).', '', ''),
        array('AlgEquiv', '', 'trigeval(sin([1/8,1/6, 1/4, 1/3, 1/2, 1]*%pi))', '[sqrt(2-sqrt(2))/2,1/2,1/sqrt(2),sqrt(3)/2,1,0]',
            1, '', '', ''),

        array('AlgEquiv', '', 'log(a^2*b)', '2*log(a)+log(b)', 1, '', 'Logarithms'),
        array('AlgEquiv', '', '(2*log(2*x)+x)/(2*x)', '(log(2*x)+2)/(2*sqrt(x))', 0, '', ''),
        array('AlgEquiv', '', 'log(abs((x^2-9)))', 'log(abs(x-3))+log(abs(x+3))', 0, '', ''),
        array('AlgEquiv', '', 'lg(10^x)', 'x', 1, '', ''),
        array('AlgEquiv', '', 'lg(3^x,3)', 'x', 1, '', ''),
        array('AlgEquiv', '', 'lg(a^x,a)', 'x', 1, '', ''),
        array('AlgEquiv', '', '1+lg(27,3)', '4', 1, '', ''),
        array('AlgEquiv', '', '1+lg(27,3)', '3', 0, '', ''),
        array('AlgEquiv', '', 'lg(1/8,2)', '-3', 1, '', ''),
        array('AlgEquiv', '', 'lg(root(x,n))', 'lg(x,10)/n', 1, '', ''),
        // The log(x) function is base e.
        array('AlgEquiv', '', 'log(root(x,n))', 'lg(x,10)/n', 0, '', ''),
        array('AlgEquiv', '', 'x^log(y)', 'y^log(x)', 1, '', ''),
        array('AlgEquiv', '', 'e^1-e^(-1)', '2*sinh(1)', 1, '', 'Hyperbolic trig'),
        array('AlgEquiv', '', 'x', '[1,2,3]', 0, 'ATAlgEquiv_SA_not_list.', 'Lists'),
        array('AlgEquiv', '', '[1,2]', '[1,2,3]', 0, 'ATList_wronglen.', ''),
        array('AlgEquiv', '', '[1,2,4]', '[1,2,3]', 0, '(ATList_wrongentries 3).', ''),
        array('AlgEquiv', '', '[1,x>2]', '[1,2<x]', 1, '', ''),
        array('AlgEquiv', '', '[1,2,[2-x<0,{1,2,2,2, 1,3}]]', '[1,2,[2-x<0,{1,2}]]', 0,
            '(ATList_wrongentries 3: (ATList_wrongentries 2: ATSet_wrongsz)).', ''),
        array('AlgEquiv', '', '[(k+8)/(k^2+4*k-12),-(2*k+6)/(k^2+4*k-12)]', '[(k+8)/(k^2+4*k-12),-(2*k+6)/(k^2+4*k-12)]',
            1, '', ''),

        // Note to self: Maxima's round() command uses Bankers' rounding, but significantfigures does not.
        array('AlgEquiv', '', 'round(0.5)', '0.0', 1, '', 'Rounding of floats'),
        array('AlgEquiv', '', 'round(1.5)', '2.0', 1, '', ''),
        array('AlgEquiv', '', 'round(2.5)', '2.0', 1, '', ''),
        array('AlgEquiv', '', 'round(12.5)', '12.0', 1, '', ''),
        array('AlgEquiv', '', 'significantfigures(0.5,1)', '0.5', 1, '', ''),
        array('AlgEquiv', '', 'significantfigures(1.5,1)', '2.0', 1, '', ''),
        array('AlgEquiv', '', 'significantfigures(2.5,1)', '3.0', 1, '', ''),
        array('AlgEquiv', '', 'significantfigures(3.5,1)', '4.0', 1, '', ''),
        array('AlgEquiv', '', 'significantfigures(11.5,2)', '12.0', 1, '', ''),
        array('AlgEquiv', '', '[3,3.1,3.14,3.142,3.1416,3.14159,3.141593,3.1415927]',
            'makelist(significantfigures(%pi,i),i,8)', 1, '', ''),
        array('AlgEquiv', '', 'x', '{1,2,3}', 0, 'ATAlgEquiv_SA_not_set.', 'Sets'),
        array('AlgEquiv', '', 'co(1,2)', '{1,2,3}', 0, 'ATAlgEquiv_SA_not_set.', ''),
        array('AlgEquiv', '', '{1,2}', '{1,2,3}', 0, 'ATSet_wrongsz.', ''),
        array('AlgEquiv', '', '{2/4, 1/3}', '{1/2, 1/3}', 1, '', ''),
        array('AlgEquiv', '', '{A[1],A[2],A[4]}', '{A[1],A[2],A[3]}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{A[1],A[2],A[3]}', '{A[1],A[2],A[3]}', 1, '', ''),
        array('AlgEquiv', '', '{1,2,4}', '{1,2,3}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{1,x>4}', '{4<x, 1}', 1, '', ''),
        array('AlgEquiv', '', '{x-1=0,x>1 and 5>x}', '{x>1 and x<5,x=1}', 1, '', ''),
        array('AlgEquiv', '', '{x-1=0,x>1 and 5>x}', '{x>1 and x<5,x=2}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{x-1=0,x>1 and 5>x}', '{x>1 and x<3,x=1}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{-sqrt(2)/sqrt(3)}', '{-2/sqrt(6)}', -3, 'ATSet_wrongentries.',
            'Equivalence for elements of sets is different from expressions: see docs.'),
        array('AlgEquiv', '', '{[-sqrt(2)/sqrt(3),0],[2/sqrt(6),0]}', '{[2/sqrt(6),0],[-2/sqrt(6),0]}', -3,
            'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', 'ev(radcan({-sqrt(2)/sqrt(3)}),simp)', 'ev(radcan({-2/sqrt(6)}),simp)', 1, '', ''),
        array('AlgEquiv', '', 'ev(radcan(ratsimp({(-sqrt(10)/2)-2,sqrt(10)/2-2},algebraic:true)),simp)',
            'ev(radcan(ratsimp({(-sqrt(5)/sqrt(2))-2,sqrt(5)/sqrt(2)-2},algebraic:true)),simp)', 1, '', ''),
        // We don't simplify here.
        array('AlgEquiv', '', '{(2-2^(5/2))/2,(2^(5/2)+2)/2}', '{1-2^(3/2),2^(3/2)+1}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', 'ev(radcan({(2-2^(5/2))/2,(2^(5/2)+2)/2}),simp)', '{1-2^(3/2),2^(3/2)+1}', 1, '', ''),
        array('AlgEquiv', '', '{(x-a)^6000}', '{(a-x)^6000}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{(k+8)/(k^2+4*k-12),-(2*k+6)/(k^2+4*k-12)}', '{(k+8)/(k^2+4*k-12),-(2*k+6)/(k^2+4*k-12)}',
            1, '', ''),

        array('AlgEquiv', '', 'matrix([1,2],[2,3])', 'matrix([1,2],[2,3])', 1, '', 'Matrices'),
        array('AlgEquiv', '', 'matrix([1,2],[2,3])', 'matrix([1,2,3],[2,3,3])', 0, 'ATMatrix_wrongsz_columns.', ''),
        array('AlgEquiv', '', 'matrix([1,2],[2,3])', 'matrix([1,2],[2,5])', 0, 'ATMatrix_wrongentries.', ''),
        array('AlgEquiv', '', 'matrix([0.33,1],[1,1])', 'matrix([0.333,1],[1,1])', 0, 'ATMatrix_wrongentries.', ''),
        array('AlgEquiv', '', 'matrix([x+x,2],[2,x*x])', 'matrix([2*x,2],[2,x^2])', 1, '', ''),
        array('AlgEquiv', '', 'matrix([epsilon[0],2],[2,x^2])', 'matrix([epsilon[0],2],[2,x^2])', 1, '', ''),
        array('AlgEquiv', '', 'matrix([epsilon[2],2],[2,x^2])', 'matrix([epsilon[0],2],[2,x^3])', 0, 'ATMatrix_wrongentries.', ''),
        array('AlgEquiv', '', 'matrix([x>4,{1,x^2}],[[1,2],[1,3]])', 'matrix([4-x<0,{x^2, 1}],[[1,2],[1,3]])', 1, '', ''),
        array('AlgEquiv', '', 'matrix([x>4,{1,x^2}],[[1,2],[1,3]])', 'matrix([4-x<0,{x^2, 1}],[[1,2],[1,4]])', 0,
            'ATMatrix_wrongentries.', ''),

        // A vector and a scalar are not the same.
        array('AlgEquiv', '', 'a', 'stackvector(a)', 0, '', 'Vectors'),

        array('AlgEquiv', '', '1', 'x=1', 0, 'ATAlgEquiv_SA_not_equation.', 'Equations'),
        array('AlgEquiv', '', 'x=1', 'x=1', 1, 'ATEquation_sides', ''),
        array('AlgEquiv', '', '1=x', '1=x', 1, 'ATEquation_sides', ''),
        array('AlgEquiv', '', '1=x', 'x=1', 1, 'ATEquation_sides_op', ''),
        array('AlgEquiv', '', '1=1', '1=x', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', '1=1', 'x=1', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'x=2', 'x=1', 0, 'ATEquation_lhs_notrhs', ''),
        array('AlgEquiv', '', '2=x', 'x=1', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'x=x', 'y=y', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'x+y=1', 'y=1-x', 1, '', ''),
        array('AlgEquiv', '', '2*x+2*y=1', 'y=0.5-x', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', '1/x+1/y=2', 'y = x/(2*x-1)', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'y=sin(2*x)', 'y/2=cos(x)*sin(x)', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'y=(x-a)^6000', 'y=(x-a)^6000', 1, 'ATEquation_sides', ''),
        array('AlgEquiv', '', 'y=(x-a)^5999', 'y=(x-a)^6000', 0, 'ATEquation_lhs_notrhs', ''),
        array('AlgEquiv', '', 'y=(a-x)^6000', 'y=(x-a)^6000', 1, 'ATEquation_sides', ''),
        array('AlgEquiv', '', 'y=(a-x)^5999', 'y=(x-a)^5999', 0, 'ATEquation_lhs_notrhs', ''),
        array('AlgEquiv', '', 'y=(a-x)^59999', 'y=(x-a)^5999', 0, 'ATEquation_lhs_notrhs', ''),
        array('AlgEquiv', '', 'x+y=i', 'y=i-x', 1, '', ''),
        array('AlgEquiv', '', '(1+%i)*(x+y)=0', 'y=-x', 1, '', ''),
        array('AlgEquiv', '', 's^2*%e^(s*t)=0', 's^2=0', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'x^6000-x^6001=x^5999', 'x^5999*(1-x+x^2)=0', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'x^6000-x^6001=x^5999', 'x^5999*(1-x+x^3)=0', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', '258552*x^7*(81*x^8+1)^398', 'x^3*(x^4+1)^399', 0, '', ''),
        array('AlgEquiv', '', 'Ia*(R1+R2+R3)-Ib*R3=0', '-Ia*(R1+R2+R3)+Ib*R3=0', 1, '', ''),
        array('AlgEquiv', '', 'a=0 or b=0', 'a*b=0', 1, 'ATEquation_sides', ''),
        array('AlgEquiv', '', 'a*b=0', 'a=0 or b=0', 1, 'ATEquation_sides', ''),

        array('AlgEquiv', '', '1', 'stackeq(1)', 1, '', 'Unary Equations'),
        array('AlgEquiv', '', 'stackeq(1)', '1', 1, '', ''),
        array('AlgEquiv', '', 'stackeq(1)', '0', 0, '', ''),

        array('AlgEquiv', '', 'x=y', 'x^2=y^2', 0, 'ATEquation_default',
            'Equations: Loose/gain roots with nth powers of each side.'),
        // Note that algebraic equivalence does check multiplicity of roots.
        array('AlgEquiv', '', '(x-2)^2=0', 'x=2', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'a^3*b^3=0', 'a=0 or b=0', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'a^3*b^3=0', 'a*b=0', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', '(x-y)*(x+y)=0', 'x^2=y^2', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'x=1', '(x-1)^3=0', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'sqrt(x)=sqrt(y)', 'x=y', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'x=sqrt(a)', 'x^2=a', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', '(x-sqrt(a))*(x+sqrt(a))=0', 'x^2=a', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', '(x-%i*sqrt(a))*(x+%i*sqrt(a))=0', 'x^2=-a', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', '(x-%i*sqrt(abs(a)))*(x+%i*sqrt(abs(a)))=0', 'x^2=-abs(a)', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'y=sqrt(1-x^2)', 'x^2+y^2=1', 0, 'ATEquation_default', ''),
        array('AlgEquiv', '', '(y-sqrt(1-x^2))*(y+sqrt(1-x^2))=0', 'x^2+y^2=1', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', '(y-sqrt((1-x)*(1+x)))*(y+sqrt((1-x)*(1+x)))=0', 'x^2+y^2=1', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', '(x-1)*(x+1)*(y-1)*(y+1)=0', 'y^2+x^2=1+x^2*y^2', 1, 'ATEquation_ratio', ''),

        array('AlgEquiv', '', 'all', 'x=x', 1, 'ATEquation_zero',
            'Equations: edge cases. Teacher must enter an equation, all or none here.'),
        array('AlgEquiv', '', 'true', 'x=x', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'x=x', 'all', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'all', 'all', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'true', 'all', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'a=a', 'x=x', 1, 'ATEquation_zero', ''),
        array('AlgEquiv', '', 'false', 'x=x', 0, 'ATEquation_zero_fail', ''),
        array('AlgEquiv', '', 'false', 'all', 0, 'ATEquation_zero_fail', ''),
        array('AlgEquiv', '', 'none', 'all', 0, 'ATEquation_zero_fail', ''),
        array('AlgEquiv', '', 'all', 'none', 0, 'ATEquation_empty_fail', ''),
        array('AlgEquiv', '', '2=3', '1=4', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', 'none', '1=2', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', 'false', '1=2', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', 'none', 'none', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', 'false', 'none', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', '3=0', 'none', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', '0=3', 'none', 1, 'ATEquation_empty', ''),
        array('AlgEquiv', '', 'all', '1=2', 0, 'ATEquation_empty_fail', ''),
        array('AlgEquiv', '', 'true', '1=2', 0, 'ATEquation_empty_fail', ''),
        array('AlgEquiv', '', '{}', '1=2', 0, 'ATAlgEquiv_SA_not_equation.', ''),
        array('AlgEquiv', '', '[]', '1=2', 0, 'ATAlgEquiv_SA_not_equation.', ''),
        array('AlgEquiv', '', '{}', 'none', 0, 'ATAlgEquiv_SA_not_logic.', ''),

        array('AlgEquiv', '', 'x^2', 'cc(1,3)', 0, 'ATAlgEquiv_SA_not_realset.', 'Sets of real numbers'),
        array('AlgEquiv', '', '%union(oo(1,2),oo(3,4))', '%union(oo(1,2),oo(3,4))', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', '%union(oc(1,2),co(2,3))', 'oo(1,3)', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', '%union(oc(1,2),co(2,3))', 'cc(1,3)', 0, 'ATRealSet_false.', ''),
        // Because we have bumped the "type" of the teacher's answer to realset, this test goes to ATRealSet, not sets.
        array('AlgEquiv', '', '{-1,1}', '%union({-1,1})', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', '{1,3}', 'cc(1,3)', 0, 'ATRealSet_false.', ''),
        array('AlgEquiv', '', '%intersection(oc(-1,1),co(1,2))', '%union({1})', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', 'oo(-inf,1)', 'oo(-inf,1)', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', 'oo(-1,inf)', 'oo(0,inf)', 0, 'ATRealSet_false.', ''),
        array('AlgEquiv', '', '%union(oc(-inf,0),oo(-1,4))', 'oo(-inf,4)', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', '%union(oo(-inf,1),oo(-1,inf))', 'oo(-inf,inf)', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', 'all', 'oo(-inf,inf)', 1, 'ATRealSet_true.', ''),

        array('AlgEquiv', '', 'a=b/%i', '%i*a=b', 1, 'ATEquation_num_i', 'Complex numbers'),
        array('AlgEquiv', '', 'b/%i=a', '%i*a=b', 1, 'ATEquation_num_i', ''),
        array('AlgEquiv', '', 'b=a/%i', '%i*a=b', 0, 'ATEquation_lhs_notrhs_op', ''),
        array('AlgEquiv', '', 'a*(2+%i)=b', 'a=b/(2+%i)', 1, 'ATEquation_ratio', ''),
        array('AlgEquiv', '', 'a*(2+%i)=b', 'a=b*(2-%i)/5', 1, 'ATEquation_num_i', ''),
        array('AlgEquiv', '', 'a*(2+%i)=b', 'a=b*(2-%i)/4', 0, 'ATEquation_default', ''),

        array('AlgEquiv', '', 'abs(x)=abs(y)', 'x=y', 0, 'ATEquation_default', 'Absolute value in equations'),
        array('AlgEquiv', '', 'abs(x)=abs(y)', 'x=y or x=-y', 1, '', ''),
        array('AlgEquiv', '', 'abs(x)=abs(y)', '(x-y)*(x+y)=0', 1, '', ''),

        array('AlgEquiv', '', 'f(x):=1/0', 'f(x):=x^2', -1, 'TEST_FAILED', 'Functions'),
        array('AlgEquiv', '', '1', 'f(x):=x^2', 0, 'ATAlgEquiv_SA_not_function.', ''),
        array('AlgEquiv', '', 'f(x)=x^2', 'f(x):=x^2', 0, 'ATAlgEquiv_SA_not_function.', ''),
        array('AlgEquiv', '', 'f(x):=x^2', 'f(x,y):=x^2+y^2', 0, 'ATFunction_length_args. ATFunction_false.', ''),
        array('AlgEquiv', '', 'f(x):=x^2', 'f(x)=x^2', 0, 'ATAlgEquiv_SA_not_equation.', ''),
        array('AlgEquiv', '', 'f(x):=x^2', 'f(x):=x^2', 1, 'ATFunction_true.', ''),
        array('AlgEquiv', '', 'f(x):=x^2', 'f(x):=sin(x)', 0, 'ATFunction_false.', ''),
        array('AlgEquiv', '', 'g(x):=x^2', 'f(x):=x^2', 0, 'ATFunction_wrongname. ATFunction_true.', ''),
        array('AlgEquiv', '', 'f(y):=y^2', 'f(x):=x^2', 1, 'ATFunction_arguments_different. ATFunction_true.', ''),
        array('AlgEquiv', '', 'f(a,b):=a^2+b^2', 'f(x,y):=x^2+y^2', 1, 'ATFunction_arguments_different. ATFunction_true.', ''),

        array('AlgEquiv', '', '1', 'x>1', 0, 'ATAlgEquiv_SA_not_inequality.', 'Inequalities'),
        array('AlgEquiv', '', 'x=1', 'x>1 and x<5', 0, 'ATAlgEquiv_TA_not_equation.', ''),
        array('AlgEquiv', '', 'x<1', 'x>1', 0, 'ATInequality_backwards.', ''),
        array('AlgEquiv', '', '1<x', 'x>1', 1, '', ''),
        array('AlgEquiv', '', 'a<b', 'b>a', 1, '', ''),
        array('AlgEquiv', '', '2<2*x', 'x>1', 1, '', ''),
        array('AlgEquiv', '', '-2>-2*x', 'x>1', 1, '', ''),
        array('AlgEquiv', '', 'x>1', 'x<=1', 0, 'ATInequality_strict. ATInequality_backwards.', ''),
        array('AlgEquiv', '', 'x>=2', 'x<2', 0, 'ATInequality_nonstrict. ATInequality_backwards.', ''),
        array('AlgEquiv', '', 'x>=1', 'x>2', 0, 'ATInequality_nonstrict.', ''),
        array('AlgEquiv', '', 'x>1', 'x>1', 1, '', ''),
        array('AlgEquiv', '', 'x>=1', 'x>=1', 1, '', ''),
        array('AlgEquiv', '', 'x>2', 'x>1', 0, '', ''),
        array('AlgEquiv', '', '1<x', 'x>1', 1, '', ''),
        array('AlgEquiv', '', '2*x>=x^2', 'x^2<=2*x', 1, '', ''),
        array('AlgEquiv', '', '2*x>=x^2', 'x^2<=2*x', 1, '', ''),
        array('AlgEquiv', '', '3*x^2<9*a', 'x^2-3*a<0', 1, '', ''),
        array('AlgEquiv', '', 'x^2>4', 'x>2 or x<-2', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '1<x or x<-3', 'x<-3 or 1<x', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '1<x or x<-3', 'x<-1 or 3<x', 0, '', ''),
        array('AlgEquiv', '', 'x>1 and x<5', 'x>1 and x<5', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'x>1 and x<5', '5>x and 1<x', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'not (x<=2 and -2<=x)', 'x>2 or -2>x', 1, 'ATLogic_True.', ''),

        array('AlgEquiv', '', 'x>2 or -2>x', 'not (x<=2 and -2<=x)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'x>=1 or 1<=x', 'x>=1', 1, '', ''),
        array('AlgEquiv', '', 'x>=1 and x<=1', 'x=1', 1, 'ATInequality_solver.', ''),
        array('AlgEquiv', '', '(x>4 and x<5) or (x<-4 and x>-5) or (x+5>0 and x<-4)',
            '(x>-5 and x<-4) or (x>4 and x<5)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(x>4 and x<5) or (x<-4 and x>-5) or (x+5>0 and x<-4)',
            '(x>-5 and x<-4) or (x>8 and x<5)', 0, '', ''),
        array('AlgEquiv', '', '(x < 0 nounor x >= 1) nounand x <= 3', 'x < 0 or (x >= 1 and x <= 3)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(x < 0 nounor x >= 1) nounand x <= 3', 'x < 0 or x >= 1 and x <= 3', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(x < 0 nounor x >= 1) nounand x <= 3', 'x < 0 or (x >= 1 and x <= 3)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(x < 0 nounor x >= 1) nounand x <= 3', '(x < 0 or x >= 1) and x <= 3', 1,
            'ATLogic_True.', ''),
        array('AlgEquiv', '', '(x < 0 nounor x >= 1) nounand x <= 3', 'x < 0 or (x >= 1 and x <= 3)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'natural_domain(1/x^2)', 'natural_domain(1/x)', 1, 'ATRealSet_true.', ''),
        array('AlgEquiv', '', 'x^4>=0', 'x^2>=0', 1, '', ''),
        array('AlgEquiv', '', 'x^4>=16', 'x^2>=4', 1, '', ''),
        array('AlgEquiv', '', '2*x^2+x>=6', 'x<=-2 or x>=3/2', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '{2,-2}', 'x>2 nounor -2>x', 0, 'ATAlgEquiv_SA_not_logic.', ''),
        array('AlgEquiv', '', 'x^2<4', 'x<2 nounand x>-2', 1, 'ATLogic_Solver_True.', ''),
        array('AlgEquiv', '', 'x^2<6', 'x<2 nounand x>-2', 0, '', ''),
        array('AlgEquiv', '', 'x>1 nounand x<-1', 'false', 1, 'ATLogic_Solver_True.', ''),
        array('AlgEquiv', '', 'x>1 nounand x<3', 'true', 0, '', ''),
        array('AlgEquiv', '', 'x>1 nounor x<3', 'true', 1, 'ATLogic_Solver_True.', ''),
        array('AlgEquiv', '', 'x>1 nounor x<3', 'all', 1, 'ATLogic_Solver_True.', ''),
        array('AlgEquiv', '', 'abs(x)<1', 'abs(x)<1', 1, '', ''),
        array('AlgEquiv', '', 'abs(x)<1', 'abs(x)<2', 0, '', ''),
        array('AlgEquiv', '', 'abs(x)<1', 'abs(x)>1', 0, 'ATInequality_backwards.', ''),
        array('AlgEquiv', '', 'abs(x)<2', '-2<x and x<2', -3, '', ''),
        array('AlgEquiv', '', '-2<x and x<2', 'abs(x)<2',-3, '', ''),
        array('AlgEquiv', '', 'abs(x)<2', '-1<x and x<1', 0, '', ''),
        array('AlgEquiv', '', 'x^2<=9', 'abs(x)<3', 0, '', ''),
        array('AlgEquiv', '', 'x^2<=9', 'abs(x)<=3', -3, '', ''),
        array('AlgEquiv', '', 'x^6<1', 'abs(x)<1', -3, '', ''),
        array('AlgEquiv', '', 'abs(x)>1', 'x<-1 or x>1', -3, '', ''),

        array('AlgEquiv', '', 'sqrt(12)', '2*sqrt(3)', 1, '', 'Surds'),
        array('AlgEquiv', '', 'sqrt(11+6*sqrt(2))', '3+sqrt(2)', 1, '', ''),
        array('AlgEquiv', '', '(19601-13860*sqrt(2))^(7/4)', '(5*sqrt(2)-7)^7', 1, '', ''),
        array('AlgEquiv', '', '(19601-13861*sqrt(2))^(7/4)', '(5*sqrt(2)-7)^7', 0, '', ''),
        array('AlgEquiv', '', '(19601-13861*sqrt(2))^(7/4)', '(5*sqrt(2)-7)^7', 0, '', ''),
        array('AlgEquiv', '', 'sqrt(2*log(26)+4-2*log(2))', 'sqrt(2*log(13)+4)', 1, '', ''),
        array('AlgEquiv', '', 'sqrt(2)*sqrt(3)+2*(sqrt(2/3))*x-(2/3)*(sqrt(2/3))*x^2+(4/9)*(sqrt(2/3))*x^3',
            '4*sqrt(6)*x^3/27-(2*sqrt(6)*x^2)/9+(2*sqrt(6)*x)/3+sqrt(6)', 1, '', ''),
        array('AlgEquiv', '', '(n+1)*n!', '(n+1)!', 1, '', 'Factorials and binomials'),
        array('AlgEquiv', '', 'n/n!', '1/(n-1)!', 1, '', ''),
        array('AlgEquiv', '', 'n/n!', '1/(n+1)!', 0, '', ''),
        array('AlgEquiv', '', 'n!/(k!*(n-k)!)', 'binomial(n,k)', 1, '', ''),
        array('AlgEquiv', '', 'binomial(n,k)+binomial(n,k+1)', 'binomial(n+1,k+1)', -3, '', ''),
        array('AlgEquiv', '', 'binomial(n,k)+binomial(n,k+1)', 'binomial(n+1,k)', 0, '', ''),
        array('AlgEquiv', '', 'binomial(n,k)', 'binomial(n,n-k)', 1, '', ''),
        array('AlgEquiv', '', '175!*56!/(55!*176!)', '17556/55176', 1, '', ''),
        array('AlgEquiv', '', '3*s*diff(q(s),s)', '3*s*diff(q(s),s)', 1, '', 'Unevaluated derviatives'),
        array('AlgEquiv', '', 'sum(k^n,n,0,3)', 'sum(k^n,n,0,3)', 1, '', 'Sums and products'),
        array('AlgEquiv', '', '1+k+k^2+k^3', 'sum(k^n,n,0,3)', 1, '', ''),
        array('AlgEquiv', '', '1+k+k^2', 'sum(k^n,n,0,3)', 0, '', ''),
        array('AlgEquiv', '', 'product(cos(k*x),k,1,3)', 'product(cos(k*x),k,1,3)', 1, '', ''),
        array('AlgEquiv', '', 'cos(x)*cos(2*x)*cos(3*x)', 'product(cos(k*x),k,1,3)', 1, '', ''),
        array('AlgEquiv', '', 'cos(x)*cos(2*x)', 'product(cos(k*x),k,1,3)', 0, '', ''),
        array('AlgEquiv', '', '9.81*m/s^2', 'stackunits(9.81,m/s^2)', 1, '', 'Scientific units are ignored'),
        array('AlgEquiv', '', '6*stackunits(1,m)', 'stackunits(6,m)', 1, '', ''),
        array('AlgEquiv', '', 'stackunits(2,m)^2', 'stackunits(4,m^2)', 1, '', ''),
        array('AlgEquiv', '', 'stackunits(2,s)^2', 'stackunits(4,m^2)', 0, '', ''),
        array('AlgEquiv', '', '-inf', 'minf', 0, '', 'Maxima does not simplify -inf (I agree!)'),
        array('AlgEquiv', '', '2/%i*ln(sqrt((1+z)/2)+%i*sqrt((1-z)/2))', '-%i*ln(z+%i*sqrt(1-z^2))', -3,
            '', 'These currently fail'),
        array('AlgEquiv', '', 'abs(x^2-4)/(abs(x-2)*abs(x+2))', '1', -3, '', ''),
        array('AlgEquiv', '', 'abs(x^2-4)', 'abs(x-2)*abs(x+2)', -3, '', ''),
        array('AlgEquiv', '', '(-1)^n*cos(x)^n', '(-cos(x))^n', -3, '', ''),
        array('AlgEquiv', '', '(sqrt(108)+10)^(1/3)-(sqrt(108)-10)^(1/3)', '2', -3, '', ''),
        array('AlgEquiv', '', '(sqrt(2+sqrt(2))+sqrt(2-sqrt(2)))/(2*sqrt(2))', 'sqrt(sqrt(2)+2)/2', -3, '', ''),
        array('AlgEquiv', '', 'sqrt(2*x*sqrt(x^2+1)+2*x^2+1)-sqrt(x^2+1)-x', '0', -3, '', ''),
        array('AlgEquiv', '', '(77+20*sqrt(13))^(1/6)-(77-20*sqrt(13))^(1/6)', '1', -3, '', ''),
        array('AlgEquiv', '', '(930249+416020*sqrt(5))^(1/30)-(930249-416020*sqrt(5))^(1/30)', '1', -3, '', ''),
        array('AlgEquiv', '', '14336000000*x^13+250265600000*x^12+1862860800000*x^11+7623925760000*x^10+' .
            '18290677760000*x^9+24744757985280*x^8+14567212351488*x^7-3267871272960*x^6-6408053107200*x^5+' .
            '670406720000*x^4+1179708800000*x^3-429244800000*x^2+56696000000*x-2680000000',
            '512*(2*x+5)^7*(5*x-1)^5*(70*x+67)', 1, '', ''),
        array('AlgEquiv', '', '14336000000*x^13+250265600000*x^12+1862860800000*x^11+7623925760000*x^10+' .
            '18290677760000*x^9+24744757985280*x^8+14567212351488*x^7-3267871272960*x^6-6408053107200*x^5+' .
            '670406720000*x^4+1179708800000*x^3-429244800000*x^2+56696000000*x-2680000001',
            '512*(2*x+5)^7*(5*x-1)^5*(70*x+67)', 0, '', ''),
        array('AlgEquiv', '', '14336000000*x^13', '512*(2*x+5)^7*(5*x-1)^5*(70*x+67)', 0, '', ''),
        array('AlgEquiv', '', 'a*(1+sqrt(2))=b', 'a=b*(sqrt(2)-1)/3', -3, 'ATEquation_default', ''),
        array('AlgEquiv', '', 'atan(1/2)', '%pi/2-atan(2)', -3, '', 'This is only equivalent for x>=0...', ''),
        array('AlgEquiv', '', 'asinh(x)', 'ln(x+sqrt(x^2+1))', -3, '', 'This is true for all x...', ''),
        // An example due to Gauss.  Just for fun!
        array('AlgEquiv', '', 'cos(2*%pi/17)', '(-1+sqrt(17)+sqrt(34-2*sqrt(17)))/16+' .
            '(2*sqrt(17+3*sqrt(17)-sqrt(34-2*sqrt(17))-2*sqrt(34+2*sqrt(17))))/16', -3, '', '', ''),

        array('AlgEquiv', '', 'true and false', 'false', 1, 'ATLogic_True.', 'Logical expressions'),
        array('AlgEquiv', '', 'true or false', 'false', 0, '', ''),
        array('AlgEquiv', '', 'A and B', 'B and A', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'A and B', 'C and A', 0, '', ''),
        array('AlgEquiv', '', 'A and B=C', 'C=B and A', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'A and (B and C)', 'A and B and C', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'A and (B or C)', 'A and (B or C)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(A and B) or (A and C)', 'A and (B or C)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'x=-b#pm#c^2', 'x=c^2-b or x=-c^2-b', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'not(A) and not(B)', 'not(A or B)', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'not(A) and not(B)', 'not(A and B)', 0, '', ''),
        array('AlgEquiv', '', 'not(A) or B', 'boolean_form(A implies B)', 1, '', ''),
        array('AlgEquiv', '', 'not(A) or B', 'A implies B', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', 'not(A) and B', 'A implies B', 0, '', ''),
        array('AlgEquiv', '', '(not A and B) or (not B and A)', 'A xor B', 1, 'ATLogic_True.', ''),
        array('AlgEquiv', '', '(A and B) or (not A and not B)', 'A xnor B', 1, 'ATLogic_True.', ''),
        // We can't apply this simplification to sets, as it breaks sets of inequalities.
        array('AlgEquiv', '', '{not(A) or B,A and B}', '{A implies B,A and B}', 0, 'ATSet_wrongentries.', ''),
        array('AlgEquiv', '', '{A implies B,A and B}', '{not(A) and B,A and B}', 0, 'ATSet_wrongentries.', ''),

        // Algebraic equivalence evaluates nouns.
        array('AlgEquiv', '', 'diff(x^2,x)', '2*x', 1, '', 'Differential equations'),
        array('AlgEquiv', '', 'diff(x^2,x)', '\'diff(x^2,x)', 1, '', ''),
        array('AlgEquiv', '', 'noundiff(x^2,x)', '2*x', 1, '', ''),
        array('AlgEquiv', '', 'diff(y,x)', '0', 1, '', ''),
        array('AlgEquiv', '', 'noundiff(y,x)', '0', 1, '', ''),
        // Note evaluated functions.
        array('AlgEquiv', '', 'diff(y(x),x)', '0', 0, '', ''),

        array('AlgEquiv', '', '"Hello"', '"Hello"', 1, 'ATAlgEquiv_String', 'Basic support for strings'),
        array('AlgEquiv', '', '"hello"', '"Hello"', 0, 'ATAlgEquiv_String', ''),
        array('AlgEquiv', '', 'W', '"Hello"', 0, 'ATAlgEquiv_SA_not_string.', ''),
        array('AlgEquiv', '', '"Hello"', 'x^2', 0, 'ATAlgEquiv_SA_not_expression.', ''),

        array('AlgEquivNouns', '', '1/0', '1', -1, 'ATAlgEquivNouns_STACKERROR_SAns.', ''),
        array('AlgEquivNouns', '', '1', '1/0', -1, 'ATAlgEquivNouns_STACKERROR_TAns.', ''),
        array('AlgEquivNouns', '', '', '(x-1)^2', -1, 'ATAlgEquivNounsTEST_FAILED-Empty SA.', ''),
        array('AlgEquivNouns', '', 'x^2', '', -1, 'ATAlgEquivNounsTEST_FAILED-Empty TA.', ''),
        array('AlgEquivNouns', '', 'x-1)^2', '(x-1)^2', -1, 'ATAlgEquivNounsTEST_FAILED-Empty SA.', ''),

        array('AlgEquivNouns', '', 'diff(x^2,x)', '2*x', 1, '', ''),
        array('AlgEquivNouns', '', 'diff(x^2,x)', '\'diff(x^2,x)', 0, '', ''),
        array('AlgEquivNouns', '', 'diff(x^2,x)', '\'diff(x^2,x)', 0, '', ''),
        array('AlgEquivNouns', '', '\'diff(y,x)', 'noundiff(y,x)', 1, '', ''),
        array('AlgEquivNouns', '', 'diff(y,x)', '0', 1, '', ''),
        array('AlgEquivNouns', '', '\'diff(y,x)', '0', 0, '', ''),
        array('AlgEquivNouns', '', 'noundiff(y,x)', '0', 0, '', ''),
        array('AlgEquivNouns', '', 'diff(y(x),x)', '0', 0, '', ''),
        array('AlgEquivNouns', '', 'noundiff(H,x,2) = -R/T', 'noundiff(H,x,2) + R/T = 0', 1, 'ATEquation_ratio',
                'Differential equations'),
        array('AlgEquivNouns', '', '\'diff(H,x,2) = -R/T', 'noundiff(H,x,2) + R/T = 0', 1, 'ATEquation_ratio', ''),
        array('AlgEquivNouns', '', 'y(t)=int(s^2,s,0,t)', 'y(t)=t^3/3', 1, 'ATEquation_sides', ''),
        array('AlgEquivNouns', '', 'y(t)=\'int(s^2,s,0,t)', 'y(t)=t^3/3', 0, 'ATEquation_lhs_notrhs', ''),
        array('AlgEquivNouns', '', 'y(t)=\'int(s^2,s,0,t)', 'y(t)=nounint(s^2,s,0,t)', 1, 'ATEquation_sides', ''),
        array('AlgEquivNouns', '', 'true nounand false', 'false', 1, 'ATLogic_True.',
                'Logic nouns are still evaluated'),

        array('SubstEquiv', '', '1/0', 'x^2-2*x+1', -1, 'ATSubstEquiv_STACKERROR_SAns.', ''),
        array('SubstEquiv', '[1/0]', 'x^2', 'x^2-2*x+1', -1, 'ATSubstEquiv_STACKERROR_Opt.', ''),
        array('SubstEquiv', 'x', 'x^2', 'x^2-2*x+1', -1, 'ATSubstEquiv_Opt_List.', ''),
        array('SubstEquiv', '', 'x^2+1', 'x^2+1', 1, '', ''),
        array('SubstEquiv', '', 'x^2+1', 'x^3+1', 0, '', ''),
        array('SubstEquiv', '', 'x^2+1', 'x^3+1', 0, '', ''),
        array('SubstEquiv', '', 'X^2+1', 'x^2+1', 1, 'ATSubstEquiv_Subst [X = x].', ''),
        array('SubstEquiv', '', 'x^2+y', 'a^2+b', 1, 'ATSubstEquiv_Subst [x = a,y = b].', ''),
        array('SubstEquiv', '', 'x^2+y/z', 'a^2+c/b', 1, 'ATSubstEquiv_Subst [x = a,y = c,z = b].', ''),
        array('SubstEquiv', '', 'y=x^2', 'a^2=b', 1, 'ATSubstEquiv_Subst [x = a,y = b].', ''),
        array('SubstEquiv', '', '{x=1,y=2}', '{x=2,y=1}', 1, 'ATSubstEquiv_Subst [x = y,y = x].', ''),
        array('SubstEquiv', '', 'cos(a*x)/(x*(ln(x)))', 'cos(a*y)/(y*(ln(y)))', 1, 'ATSubstEquiv_Subst [a = a,x = y].',
                'Where a variable is also a function name.'),
        array('SubstEquiv', '', 'cos(a*x)/(x*(ln(x)))', 'cos(x*a)/(a*(ln(a)))', 1, 'ATSubstEquiv_Subst [a = x,x = a].', ''),
        array('SubstEquiv', '', 'cos(a*x)/(x*(ln(x)))', 'cos(a*x)/(x(ln(x)))', 0, '', ''),
        array('SubstEquiv', '', 'cos(a*x)/(x*(ln(x)))', 'cos(a*y)/(y(ln(y)))', 0, '', ''),
        array('SubstEquiv', '', 'x+1>y', 'y+1>x', 1, 'ATSubstEquiv_Subst [x = y,y = x].', ''),
        array('SubstEquiv', '', 'x+1>y', 'x<y+1', 1, 'ATSubstEquiv_Subst [x = y,y = x].', ''),
        array('SubstEquiv', '', 'matrix([1,A^2+A+1],[2,0])', 'matrix([1,x^2+x+1],[2,0])', 1,
            'ATSubstEquiv_Subst [A = x].', 'Matrices'),
        array('SubstEquiv', '', 'matrix([B,A^2+A+1],[2,C])', 'matrix([y,x^2+x+1],[2,z])', 1,
            'ATSubstEquiv_Subst [A = x,B = y,C = z].', ''),
        array('SubstEquiv', '', 'matrix([B,A^2+A+1],[2,C])', 'matrix([y,x^2+x+1],[2,x])', 0, 'ATMatrix_wrongentries.', ''),
        array('SubstEquiv', '', '[x^2+1,x^2]', '[A^2+1,A^2]', 1, 'ATSubstEquiv_Subst [x = A].', 'Lists'),
        array('SubstEquiv', '', '[x^2-1,x^2]', '[A^2+1,A^2]', 0, '(ATList_wrongentries 1, 2).', ''),
        array('SubstEquiv', '', '[A,B,C]', '[B,C,A]', 1, 'ATSubstEquiv_Subst [A = B,B = C,C = A].', ''),
        array('SubstEquiv', '', '[A,B,C]', '[B,B,A]', 0, '(ATList_wrongentries 1, 3).', ''),
        array('SubstEquiv', '', '[1,[A,B],C]', '[1,[a,b],C]', 1, 'ATSubstEquiv_Subst [A = a,B = b,C = C].', ''),
        array('SubstEquiv', '', '{x^2+1,x^2}', '{A^2+1,A^2}', 1, 'ATSubstEquiv_Subst [x = A].', 'Sets'),
        array('SubstEquiv', '', '{x^2-1,x^2}', '{A^2+1,A^2}', 0, 'ATSet_wrongentries.', ''),
        array('SubstEquiv', '', '{A+1,B^2,C}', '{B,C+1,A^2}', 1, 'ATSubstEquiv_Subst [A = C,B = A,C = B].', ''),
        array('SubstEquiv', '', '{1,{A,B},C}', '{1,{a,b},C}', 1, 'ATSubstEquiv_Subst [A = a,B = b,C = C].', ''),
        // Optional argument to fix some variables within an expression.
        array('SubstEquiv', '', 'A*cos(t)+B*sin(t)', 'P*cos(t)+Q*sin(t)', 1, 'ATSubstEquiv_Subst [A = P,B = Q,t = t].', ''),
        array('SubstEquiv', '', 'A*cos(t)+B*sin(t)', 'P*cos(x)+Q*sin(x)', 1, 'ATSubstEquiv_Subst [A = P,B = Q,t = x].', ''),
        array('SubstEquiv', '[x]', 'A*cos(t)+B*sin(t)', 'P*cos(x)+Q*sin(x)', 0, '', 'Fix some variables.'),
        // Fixes variables only in the teacher's answer.
        array('SubstEquiv', '[t]', 'A*cos(t)+B*sin(t)', 'P*cos(x)+Q*sin(x)', 1, 'ATSubstEquiv_Subst [A = P,B = Q,t = x].', ''),
        array('SubstEquiv', '[x,t]', 'A*cos(t)*e^x+B*sin(t)*e^x+C*sin(2*x)+D*cos(2*x)',
            'P*cos(t)*e^x+Q*sin(t)*e^x+R*sin(2*x)+S*cos(2*x)', 1, 'ATSubstEquiv_Subst [A = P,B = Q,C = R,D = S].', ''),

        array('EqualComAss', '', '1/0', '0', -1, 'ATEqualComAss_STACKERROR_SAns.', ''),
        array('EqualComAss', '', '0', '1/0', -1, 'ATEqualComAss_STACKERROR_TAns.', ''),
        array('EqualComAss', '', '2/4', '1/2', 0, 'ATEqualComAss (AlgEquiv-true).', 'Numbers'),
        array('EqualComAss', '', '3^2', '8', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', '3^2', '9', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'cos(0)', '1', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '4^(1/2)', '2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/3^(1/2)', '(1/3)^(1/2)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'sqrt(3)/3', '(1/3)^(1/2)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'sqrt(3)', '3^(1/2)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '2*sqrt(2)', 'sqrt(8)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '2*2^(1/2)', 'sqrt(8)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'sqrt(2)/4', '1/sqrt(8)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/sqrt(2)', '2^(1/2)/2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '4.0', '4', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'X', 'x', 0, 'ATEqualComAss (AlgEquiv-false)ATAlgEquiv_WrongCase.', 'Case sensitivity'),
        array('EqualComAss', '', 'exdowncase(X)', 'x', 1, '', ''),
        array('EqualComAss', '', 'exdowncase((X-1)^2)', 'x^2-2*x+1', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'exdowncase(X^2-2*X+1)', 'x^2-2*x+1', 1, '', ''),
        array('EqualComAss', '', 'a^2/b^3', 'a^2*b^(-3)', 0, 'ATEqualComAss (AlgEquiv-true).', 'Powers'),
        array('EqualComAss', '', 'lg(a^x,a)', 'x', 0,
            'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'x^(2/4)', 'x^(1/2)', 0,
            'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1+2*x', 'x*2+1', 1, '', 'Simple polynomials'),
        array('EqualComAss', '', '1+x', '2*x+1', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', '1+x+x', '2*x+1', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(x+y)+z', 'z+x+y', 1, '', ''),
        array('EqualComAss', '', 'x*x', 'x^2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(x+5)*x', 'x*(5+x)', 1, '', ''),
        array('EqualComAss', '', 'x*(x+5)', '5*x+x^2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(1-x)^2', '(x-1)^2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(a-x)^6000', '(x-a)^6000', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 1, '',
            'Expressions with subscripts'),
        array('EqualComAss', '', 'rho*z*V/(4*pi*epsilon[1]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 0,
            'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', '-1+2', '2-1', 1, '', 'Unary minus'),
        array('EqualComAss', '', '-1*2+3*4', '3*4-1*2', 1, '', ''),
        array('EqualComAss', '', '(-1*2)+3*4', '10', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '-1*2+3*4', '3*4-1*2', 1, '', ''),
        array('EqualComAss', '', 'x*(-y)', '-x*y', 1, '', ''),
        array('EqualComAss', '', 'x*(-y)', '-(x*y)', 1, '', ''),
        array('EqualComAss', '', '(-x)*(-x)', 'x*x', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(-x)*(-x)', 'x^2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '-1/4*%pi*i', '-(%i*%pi)/4', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/2', '3/6', 0, 'ATEqualComAss (AlgEquiv-true).', 'Rational expressions'),
        array('EqualComAss', '', '1/(1+2*x)', '1/(2*x+1)', 1, '', ''),
        array('EqualComAss', '', '2/(4+2*x)', '1/(x+2)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(a*b)/c', 'a*(b/c)', 1, '', ''),
        // Non-trivial example of the above case.
        array('EqualComAss', '', '((x+1)/(x*(x-1)))*(x-1)', '((x+1)*(x-1))/(x*(x-1))', 1, '', ''),
        array('EqualComAss', '', '(-x)/y', '-(x/y)', 1, '', ''),
        array('EqualComAss', '', 'x/(-y)', '-(x/y)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '-1/(1-x)', '1/(x-1)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/2*1/x', '1/(2*x)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(k+8)/(k^2+4*k-12)', '(k+8)/(k^2+4*k-12)', 1, '', ''),
        array('EqualComAss', '', '(k+8)/(k^2+4*k-12)', '(k+8)/((k-2)*(k+6))', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(k+7)/(k^2+4*k-12)', '(k+8)/(k^2+4*k-12)', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', '-(2*k+6)/(k^2+4*k-12)', '-(2*k+6)/(k^2+4*k-12)', 1, '', ''),
        array('EqualComAss', '', '(a+b)/1', '(b+a)/1', 1, '', ''),
        array('EqualComAss', '', '1*x', 'x', 0, 'ATEqualComAss (AlgEquiv-true).', 'No simplicifcation here'),
        array('EqualComAss', '', '23+0*x', '23', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'x+0', 'x', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'x^1', 'x', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(1/2)*(a+b)', '(a+b)/2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/3*logbase(27,6)', 'logbase(27,6)/3', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/3*lg(27,6)', 'lg(27,6)/3', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'lg(root(x, n))', 'lg(x, 10)/n', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1/3*i', 'i/3', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '%i', 'e^(i*pi/2)', 0, 'ATEqualComAss (AlgEquiv-true).', 'Complex numbers'),
        array('EqualComAss', '', '(4*sqrt(3)*%i+4)^(1/5)', 'rectform((4*sqrt(3)*%i+4)^(1/5))', 0,
            'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(4*sqrt(3)*%i+4)^(1/5)', '8^(1/5)*(cos(%pi/15)+%i*sin(%pi/15))', 0,
            'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '(4*sqrt(3)*%i+4)^(1/5)', 'polarform((4*sqrt(3)*%i+4)^(1/5))', 0,
            'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'y=x', 'x=y', 1, '', 'Equations'),
        array('EqualComAss', '', 'x+1', 'y=2*x+1', 0, 'ATEqualComAss ATAlgEquiv_SA_not_equation.', ''),
        array('EqualComAss', '', 'y=1+2*x', 'y=2*x+1', 1, '', ''),
        array('EqualComAss', '', 'y=x+x+1', 'y=1+2*x', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'A and B', 'B and A', 1, '', 'Logic'),
        array('EqualComAss', '', 'A or B', 'B or A', 1, '', ''),
        array('EqualComAss', '', 'A or B', 'B and A', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', 'not(true)', 'false', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '{2*x+1,2}', '{2, 1+x*2}', 1, '', 'Sets'),
        array('EqualComAss', '', '2', '{2}', 0, 'ATEqualComAss ATAlgEquiv_SA_not_set.', ''),
        array('EqualComAss', '', '{2*x+1, 1+1}', '{2, 1+x*2}', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '{1,2}', '{1,{2}}', 0, 'ATEqualComAss (AlgEquiv-false)ATSet_wrongentries.', ''),
        array('EqualComAss', '', '{4,3}', '{3,4}', 1, '', ''),
        array('EqualComAss', '', '{4,4}', '{4}', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '{-1,1,-1}', '{-1,-1,1}', 1, '', ''),
        array('EqualComAss', '', '{-1,1,-1}', '{-1,1}', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '[2*x+1,2]', '[1+x*2,2]', 1, '', 'Lists'),
        array('EqualComAss', '', '[x+x+1, 1+1]', '[1+x*2,2]', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'matrix([1,2],[2,3])', 'matrix([1,2],[2,3])', 1, '', 'Matrices'),
        array('EqualComAss', '', 'matrix([1,2],[2,3])', 'matrix([1,2,3],[2,3,3])', 0,
            'ATEqualComAss (AlgEquiv-false)ATMatrix_wrongsz_columns.', ''),
        array('EqualComAss', '', 'matrix([1,2],[2,3])', 'matrix([1,2],[2,5])', 0,
            'ATEqualComAss (AlgEquiv-false)ATMatrix_wrongentries.', ''),
        array('EqualComAss', '', 'matrix([1,2],[2,2+1])', 'matrix([1,2],[2,3])', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'matrix([x+x, 1],[1, 1])', 'matrix([2*x, 1],[1, 1])', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'sum(k^n,n,0,3)', 'sum(k^n,n,0,3)', 1, '', 'Sums and products'),
        array('EqualComAss', '', '1+k+k^2+k^3', 'sum(k^n,n,0,3)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'product(cos(k*x),k,1,3)', 'product(cos(k*x),k,1,3)', 1, '', ''),
        array('EqualComAss', '', 'cos(x)*cos(2*x)*cos(3*x)', 'product(cos(k*x),k,1,3)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '-6/5 > x', 'x < -6/5', 0, 'ATEqualComAss (AlgEquiv-true).',
            'Inequalities are not commutative under this test'),
        array('EqualComAss', '', 'x<1 and -3<x', '-3<x and x<1', 1, '', ''),
        array('EqualComAss', '', '1>x and -3<x', '-3<x and x<1', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'make_less_ineq(-6/5 > x)', 'x < -6/5', 1, '', ''),
        array('EqualComAss', '', 'make_less_ineq(1>x and -3<x)', '-3<x and x<1', 1, '', ''),
        array('EqualComAss', '', 'make_less_ineq(6/3 > x)', 'x < 2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', '1', 'stackeq(1)', 1, '', 'Unary Equations'),
        array('EqualComAss', '', 'stackeq(1)', '1', 1, '', ''),
        array('EqualComAss', '', 'stackeq(1+1)', '2', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'stackeq(1)', '0', 0, 'ATEqualComAss (AlgEquiv-false).', ''),

        array('EqualComAss', '', 'lowesttermsp(1/3)', 'true', 1, '', ''),
        array('EqualComAss', '', 'lowesttermsp(2/6)', 'true', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', 'lowesttermsp(x^2/x)', 'true', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', 'lowesttermsp(-y/-x)', 'true', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', 'lowesttermsp((x^2-1)/(x-1))', 'true', 0, 'ATEqualComAss (AlgEquiv-false).', ''),
        array('EqualComAss', '', 'lowesttermsp((x^2-1)/(x+2))', 'true', 1, '', ''),

        // We can't use ATAlgEquiv with rationalized as Maxima simplified sqrt(3)/3 to 1/sqrt(3).
        array('EqualComAss', '', 'rationalized(1+sqrt(3)/3)', 'true', 1, '', 'Bad things in denominators'),
        array('EqualComAss', '', 'rationalized(1+1/sqrt(3))', '[sqrt(3)]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(1/sqrt(3))', '[sqrt(3)]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(1/sqrt(2)+i/sqrt(2))', '[sqrt(2),sqrt(2)]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(sqrt(2)/2+1/sqrt(3))', '[sqrt(3)]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(1/sqrt(2)+1/sqrt(3))', '[sqrt(2),sqrt(3)]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(1/(1+i))', '[i]', 1, '', ''),
        array('EqualComAss', '', 'rationalized(1/(1+1/root(3,2)))', '[root(3,2)]', 1, '', ''),

        // Differential equations.
        // Functions are evaluated with simp:false.
        array('EqualComAss', '', 'diff(y,x)', '0', 1, '', 'Differential Equations'),
        array('EqualComAss', '', 'diff(x^2,x)', '2*x', 1, '', ''),
        array('EqualComAss', '', 'noundiff(x^2,x)', '2*x', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'diff(y,x)', '\'diff(y,x)', 0, 'ATEqualComAss (AlgEquiv-true).', ''),
        array('EqualComAss', '', 'noundiff(y,x)', '\'diff(y,x)', 1, '', ''),
        array('EqualComAss', '', '\'diff(y(x),x)', '\'diff(y(x),x,1)', 1, '', ''),
        // Test case to illustrate why we need a new answer test.
        // These equations are not the same up to commutativity and associativity, because the algebra needed.
        array('EqualComAss', '', 'noundiff(y(x),x)=-x/4', '4*noundiff(y(x),x)+x=0', 0, 'ATEqualComAss (AlgEquiv-true).', ''),

        array('CasEqual', '', '1/0', 'x^2-2*x+1', -1, 'ATCASEqual_STACKERROR_SAns.', ''),
        array('CasEqual', '', 'x', '1/0', -1, 'ATCASEqual_STACKERROR_TAns.', ''),
        array('CasEqual', 'x', '0.5', '1/2', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'x=1', '1', 0, 'ATCASEqual ATAlgEquiv_TA_not_equation.', ''),
        array('CasEqual', '', 'a', 'A', 0, 'ATCASEqual_false.', 'Case sensitivity'),
        array('CasEqual', '', 'exdowncase(X^2-2*X+1)', 'x^2-2*x+1', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', '4^(-1/2)', '1/2', 0, 'ATCASEqual (AlgEquiv-true).', 'Numbers'),
        array('CasEqual', '', 'ev(4^(-1/2),simp)', 'ev(1/2,simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', '2^2', '4', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'a^2/b^3', 'a^2*b^(-3)', 0, 'ATCASEqual (AlgEquiv-true).', 'Powers'),
        array('CasEqual', '', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 1,
            'ATCASEqual_true.', 'Expressions with subscripts'),
        array('CasEqual', '', 'rho*z*V/(4*pi*epsilon[1]*(R^2+z^2)^(3/2))', 'rho*z*V/(4*pi*epsilon[0]*(R^2+z^2)^(3/2))', 0,
            'ATCASEqual_false.', ''),
        array('CasEqual', '', '0.5', '1/2', 0, 'ATCASEqual (AlgEquiv-true).', 'Mix of floats and rational numbers'),
        array('CasEqual', '', 'x^(1/2)', 'sqrt(x)', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'ev(x^(1/2),simp)', 'ev(sqrt(x),simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'abs(x)', 'sqrt(x^2)', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'ev(abs(x),simp)', 'ev(sqrt(x^2),simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'x-1', '(x^2-1)/(x+1)', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'x+x', '2*x', 0, 'ATCASEqual (AlgEquiv-true).', 'Polynomials and rational function'),
        array('CasEqual', '', 'ev(x+x,simp)', 'ev(2*x,simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'x+x^2', 'x^2+x', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'ev(x+x^2,simp)', 'ev(x^2+x,simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', '(x-1)^2', 'x^2-2*x+1', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', '(x-1)^(-2)', '1/(x^2-2*x+1)', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', '1/n-1/(n+1)', '1/(n*(n+1))', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'cos(x)', 'cos(-x)', 0, 'ATCASEqual (AlgEquiv-true).', 'Trig functions'),
        array('CasEqual', '', 'ev(cos(x),simp)', 'ev(cos(-x),simp)', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'cos(x)^2+sin(x)^2', '1', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', '2*cos(x)^2-1', 'cos(2*x)', 0, 'ATCASEqual (AlgEquiv-true).', ''),
        array('CasEqual', '', 'imag_numberp(2*%i)', 'true', 1, 'ATCASEqual_true.', 'Predicate function wrapper'),
        array('CasEqual', '', 'imag_numberp(%e^(%i*%pi/2))', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'imag_numberp(2)', 'false', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'imag_numberp(%e^(%pi/2))', 'false', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(3*%e^(%i*%pi/6))', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(3)', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(%e^(%i*%pi/6))', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(%e^%i)', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(%e^(%pi/6))', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(3+%i)', 'false', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(%e^(%i)/4)', 'true', 1, 'ATCASEqual_true.', ''),
        array('CasEqual', '', 'complex_exponentialp(3*exp(%i*%pi/6))', 'true', 1, 'ATCASEqual_true.', ''),
        // The below test is 0 because with simp:false, -1 is ((mminus) 1) so not an integer.
        array('CasEqual', '', 'integerp(-1)', 'true', 0, 'ATCASEqual_false.', ''),
        array('CasEqual', '', 'integerp(ev(-1,simp))', 'true', 1, 'ATCASEqual_true.', ''),

        array('SameType', '', '1/0', '1', -1, 'ATSameType_STACKERROR_SAns.', ''),
        array('SameType', '', '1', '1/0', -1, 'ATSameType_STACKERROR_TAns.', ''),
        array('SameType', '', '4^(-1/2)', '1/2', 1, '', 'Numbers'),
        array('SameType', '', 'x', '[1,2,3]', 0, '', 'Lists'),
        array('SameType', '', '[1,2]', '[1,2,3]', 1, '', ''),
        array('SameType', '', '[1,x>2]', '[1,2<x]', 1, '', ''),
        array('SameType', '', '[1,x,3]', '[1,2<x,4]', 0, '', ''),
        array('SameType', '', 'x', '{1,2,3}', 0, '', 'Sets'),
        array('SameType', '', '{1,2}', '{1,2,3}', 1, '', ''),
        array('SameType', '', 'matrix([1,2],[2,3])', 'matrix([1,2],[2,3])', 1, '', 'Matrices'),
        array('SameType', '', '[[1,2],[2,3]]', 'matrix([1,2],[2,3])', 0, '', ''),
        array('SameType', '', 'matrix([1,2],[2,3])', 'matrix([1,2,3],[2,3,3])', 1, '', ''),
        array('SameType', '', 'matrix([x>4,{1,x^2}],[[1,2],[1,3]])', 'matrix([4-x<0,{x^2, 1}],[[1,2],[1,3]])', 1, '', ''),
        array('SameType', '', 'matrix([x>4,[1,x^2]],[[1,2],[1,3]])', 'matrix([4-x<0,{x^2, 1}],[[1,2],[1,4]])', 0, '', ''),
        array('SameType', '', '1', 'x=1', 0, '', 'Equations'),
        array('SameType', '', 'x=1', 'x=1', 1, '', ''),
        array('SameType', '', '1', 'x>1', 0, '', 'Inequalities'),
        array('SameType', '', 'x>2', 'x>1', 1, '', ''),
        array('SameType', '', 'x>1', 'x>=1', 1, '', ''),
        array('SameType', '', 'x>1 and x<3', 'x>=1', 1, '', ''),
        array('SameType', '', '{x>1,x<3}', 'x>=1', 0, '', ''),
        array('SameType', '', 'sqrt(2)*sqrt(3)+2*(sqrt(2/3))*x-(2/3)*(sqrt(2/3))*x^2+(4/9)*(sqrt(2/3))*x^3',
            '4*sqrt(6)*x^3/27-(2*sqrt(6)*x^2)/9+(2*sqrt(6)*x)/3+sqrt(6)', 1, '', ''),

        array('SysEquiv', '', '1/0', '[(x-1)*(x+1)=0]', -1, 'ATSysEquiv_STACKERROR_SAns.',
            'Basic tests'),
        array('SysEquiv', '', '[(x-1)*(x+1)=0]', '1/0', -1, 'ATSysEquiv_STACKERROR_TAns.', ''),
        array('SysEquiv', '', '1', '[(x-1)*(x+1)=0]', 0, 'ATSysEquiv_SA_not_list.', ''),
        array('SysEquiv', '', '[(x-1)*(x+1)=0]', '1', 0, 'ATSysEquiv_SB_not_list.', ''),
        array('SysEquiv', '', '[1]', '[90=v*t,90=(v+5)*(t-1/4)]', 0, 'ATSysEquiv_SA_not_eq_list.', ''),
        array('SysEquiv', '', '[(x-1)*(x+1)=0]', '[1]', 0, 'ATSysEquiv_SB_not_eq_list.', ''),
        array('SysEquiv', '', '[x^2]', '[(x-1)*(x+1)=0]', 0, 'ATSysEquiv_SA_not_eq_list.', ''),
        array('SysEquiv', '', '[90=v*t^t,90=(v+5)*(t-1/4)]', '[90=v*t,90=(v+5)*(t-1/4)]', 0,
            'ATSysEquiv_SA_not_poly_eq_list.', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t-1/4)]', '[90=v*t^t,90=(v+5)*(t-1/4)]', 0,
            'ATSysEquiv_SB_not_poly_eq_list.', ''),
        array('SysEquiv', '', '[x^2=1]', '[(x-1)*(x+1)=0]', 1, '', 'Tests of equivalence'),
        array('SysEquiv', '', '[x^2+y^2=4,y=x]', '[y=x,y^2=2]', 1, '', ''),
        array('SysEquiv', '', '[x^2+y^2=2,y=x]', '[y=x,y^2=2]', 0, 'ATSysEquiv_SA_system_overdetermined.', ''),
        array('SysEquiv', '', '[x=1]', '[(x-1)*(x+1)=0,(x-1)*(x-3)=0]', 1, 'ATSysEquiv_SA_Completely_solved.', ''),
        array('SysEquiv', '', '[3*a+b-c=2, a-b+2*c=5,b+c=5]', '[a=1,b=2,c=3]', 1, '', ''),
        array('SysEquiv', '', '[a=1,b=2,c=3]', '[3*a+b-c=2, a-b+2*c=5,b+c=5]', 1, 'ATSysEquiv_SA_Completely_solved.', ''),
        array('SysEquiv', '', '[x^2=1]', '[(x-1)*(x+1)*(x-2)=0]', 0, 'ATSysEquiv_SA_system_overdetermined.', ''),
        // The solution to the next one is x=1 or y=-1, not x=1 and y=-1!
        array('SysEquiv', '', '[x=1,y=-1]', '[(x-1)*(y+1)=0]', 0, 'ATSysEquiv_SA_Not_completely_solved.', ''),
        array('SysEquiv', '', '[x=1]', '[(x-1)*(x+1)=0]', 0, 'ATSysEquiv_SA_Not_completely_solved.', ''),
        array('SysEquiv', '', '[x=1]', '[(x-1)*(x+1)*y=0]', 0, 'ATSysEquiv_SA_Not_completely_solved.', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t-1/4)]', '[90=v*t,90=(v+5)*(t-1/4)]', 1, '', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t*x-1/4)]', '[90=v*t,90=(v+5)*(t-1/4)]', 0, 'ATSysEquiv_SA_extra_variables.', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t-1/4)]', '[90=v*t,90=(v+5)*(t*x-1/4)]', 0,
            'ATSysEquiv_SA_missing_variables.', ''),
        array('SysEquiv', '', '[90=v*t]', '[90=v*t,90=(v+5)*(t-1/4)]', 0, 'ATSysEquiv_SA_system_underdetermined.', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t-1/4),90=(v+6)*(t-1/5)]', '[90=v*t,90=(v+5)*(t-1/4)]', 0,
            'ATSysEquiv_SA_system_overdetermined.', ''),
        array('SysEquiv', '', '[90=v*t,90=(v+5)*(t-1/4),90=(v+6)*(t-1/5),90=(v+7)*(t-1/4),90=(v+8)*(t-1/3)]',
            '[90=v*t,90=(v+5)*(t-1/4)]', 0, 'ATSysEquiv_SA_system_overdetermined.', ''),
        array('SysEquiv', '', '[b^2=a,a=9]', '[x^2=y,y=9]', 0, 'ATSysEquiv_SA_wrong_variables.', 'Wrong variables'),
        array('SysEquiv', '', '[x^2=4]', '[x^2=4,y=9]', 0, 'ATSysEquiv_SA_missing_variables.', ''),
        array('SysEquiv', '', '[d=90,d=v*t,d=(v+5)*(t-1/4)]', '[90=v*t,90=(v+5)*(t-1/4)]', 0, 'ATSysEquiv_SA_extra_variables.', ''),
        // If we want to condone extra explicit variables in an answer we can do so.
        array('SysEquiv', '', 'stack_eval_assignments([d=90,d=v*t,d=(v+5)*(t-1/4)])', '[90=v*t,90=(v+5)*(t-1/4)]', 1, '', ''),

        array('Sets', '', '{1/0}', '{0}', -1, 'ATSets_STACKERROR_SAns.', ''),
        array('Sets', '', '{0}', '{1/0}', -1, 'ATSets_STACKERROR_TAns.', ''),
        array('Sets', '', 'x', '{1,2,3}', 0, 'ATSets_SA_not_set.', ''),
        array('Sets', '', '{1,2}', 'x', 0, 'ATSets_SB_not_set.', ''),
        array('Sets', '', '{1,2}', '{1,2,3}', 0, 'ATSets_missingentries.', ''),
        array('Sets', '', '{1,2,4}', '{1,2}', 0, 'ATSets_wrongentries.', ''),
        // Note, in the example below the wrong entry "4" displayed in the feedback does not occur in the set.
        // The set contains only the equivalent entry 2+2.  This might be confusing.
        array('Sets', '', '{1,2,2+2}', '{1,2}', 0, 'ATSets_wrongentries.', ''),
        array('Sets', '', '{5,1,2,4}', '{1,2,3}', 0, 'ATSets_wrongentries. ATSets_missingentries.', ''),
        array('Sets', '', '{2/4, 1/3}', '{1/2, 1/3}', 1, '', ''),
        array('Sets', '', '{1,2,1}', '{1,2}', 1, 'ATSets_duplicates.', 'Duplicate entries'),
        array('Sets', '', '{1,2,1+1}', '{1,2}', 1, 'ATSets_duplicates.', ''),
        array('Sets', '', '{1,2,1+1}', '{1,2,3}', 0, 'ATSets_duplicates. ATSets_missingentries.', ''),
        // We accept these are "different" as we can't simplify this without expanding.
        array('Sets', '', '{(x-a)^6000}', '{(a-x)^6000}', 0, 'ATSets_wrongentries. ATSets_missingentries.', ''),

        array('Expanded', '', '1/0', '0', -1, 'ATExpanded_STACKERROR_SAns.', ''),
        array('Expanded', '', 'x>2', 'x^2-2*x+1', 0, 'ATExpanded_SA_not_expression.', ''),
        array('Expanded', '', 'x^2-1', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', '2*(x-1)', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', '(x-1)*(x+1)', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', '(x-a)*(x-b)', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', 'x^2-(a+b)*x+a*b', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', 'x^2-a*x-b*x+a*b', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', 'cos(2*x)', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', 'p+1', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', '(p+1)*(p-1)', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', '3+2*sqrt(3)', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', '3+sqrt(12)', '0', 1, 'ATExpanded_TRUE.', ''),
        array('Expanded', '', '(1+sqrt(5))*(1-sqrt(3))', '0', 0, 'ATExpanded_FALSE.', ''),
        array('Expanded', '', '(a-x)^6000', '0', -2, 'ATExpanded_TRUE.',
            'This fails, but you are never going to ask students to do this anyway...'),

        array('FacForm', 'x', '1/0', '0', -1, 'ATFacForm_STACKERROR_SAns.', ''),
        array('FacForm', 'x', '0', '1/0', -1, 'ATFacForm_STACKERROR_TAns.', ''),
        array('FacForm', '1/0', '0', '0', -1, 'ATFacForm_STACKERROR_Opt.', ''),
        array('FacForm', 'x', '2', '2', 1, 'ATFacForm_int_true.', 'Trivial cases'),
        array('FacForm', 'x', '6', '6', 1, 'ATFacForm_int_true.', ''),
        array('FacForm', 'x', '1/3', '1/3', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '3*x^2', '3*x^2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '4*x^2', '4*x^2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '2*(x-1)', '2*x-2', 1, 'ATFacForm_true.', 'Linear integer factors'),
        array('FacForm', 'x', '2*x-2', '2*x-2', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '2*(x+1)', '2*x-2', 0, 'ATFacForm_isfactored. ATFacForm_default_true. ATFacForm_notalgequiv.', ''),
        array('FacForm', 'x', '2*x+2', '2*x-2', 0, 'ATFacForm_notfactored. ATFacForm_notalgequiv.', ''),
        array('FacForm', 'x', '2*(x+0.5)', '2*x+1', 1, 'ATFacForm_default_true.', ''),
        array('FacForm', 'x', 't*(2*x+1)', 't*(2*x+1)', 1, 'ATFacForm_true.', 'Linear factors'),
        array('FacForm', 'x', 't*x+t', 't*(x+1)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '2*x*(x-3)', '2*x^2-6*x', 1, 'ATFacForm_true.', 'Quadratic, with no const'),
        array('FacForm', 'x', '2*(x^2-3*x)', '2*x*(x-3)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', 'x*(2*x-6)', '2*x*(x-3)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(x+2)*(x+3)', '(x+2)*(x+3)', 1, 'ATFacForm_true.', 'Quadratic'),
        array('FacForm', 'x', '(x+2)*(2*x+6)', '2*(x+2)*(x+3)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(z*x+z)*(2*x+6)', '2*z*(x+1)*(x+3)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(x+t)*(x-t)', 'x^2-t^2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 't', 't^2-1', '(t-1)*(t+1)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 't', 't^2+1', 't^2+1', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'v', 'v^2+1', 'v^2+1', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'v', 'v^2-1', 'v^2-1', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'v', '-(3*w-4*v+9*u)*(3*w+4*v-u)', '-(3*w-4*v+9*u)*(3*w+4*v-u)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'k', '-(6*k*(4*b-k-1))', '6*k*(1+k-4*b)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'a', '-(6*a*(4*b-a-1))', '6*a*(1+a-4*b)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'a', '-(6*a*(4*b-a-1))', '6*a*(-(4*b)+a+1)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', 'x*(x-4+4/x)', 'x^2-4*x+4', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(2-x)*(3-x)', '(x-2)*(x-3)', 1, 'ATFacForm_true.', 'These are delicate cases!'),
        array('FacForm', 'x', '(1-x)^2', '(x-1)^2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '-(1-x)^2', '-(x-1)^2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '4*(1-x/2)^2', '(x-2)^2', 1, 'ATFacForm_default_true.', ''),
        array('FacForm', 'x', '(x-1)*(x^2+x+1)', 'x^3-1', 1, 'ATFacForm_true.', 'Cubics'),
        array('FacForm', 'x', 'x^3-x+1', 'x^3-x+1', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '7*x^3-7*x+7', '7*(x^3-x+1)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(1-x)*(2-x)*(3-x)', '-x^3+6*x^2-11*x+6', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(2-x)*(2-x)*(3-x)', '-x^3+7*x^2-16*x+12', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(2-x)^2*(3-x)', '-x^3+7*x^2-16*x+12', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(x^2-4*x+4)*(3-x)', '-x^3+7*x^2-16*x+12', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(x^2-3*x+2)*(3-x)', '-x^3+6*x^2-11*x+6', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'y', '3*y^3-6*y^2-24*y', '3*(y-4)*y*(y+2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'y', '3*(y^3-2*y^2-8*y)', '3*(y-4)*y*(y+2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'y', '3*y*(y^2-2*y-8)', '3*(y-4)*y*(y+2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'y', '3*(y^2-4*y)*(y+2)', '3*(y-4)*y*(y+2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'y', '(y-4)*y*(3*y+6)', '3*(y-4)*y*(y+2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '(a-x)^6000', '(a-x)^6000', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(x-a)^6000', '(a-x)^6000', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'a', '2*a*(a*b-1)', '2*a*(a*b-1)', 1, 'ATFacForm_true.', 'Needs flattening'),
        array('FacForm', 'a', '(2*a)*(a*b-1)', '2*a*(a*b-1)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'sin(x)', '(sin(x)+1)*(sin(x)-1)', 'sin(x)^2-1', 1, 'ATFacForm_true.', 'Not polynomials in a variable'),
        array('FacForm', 'cos(t)', '(cos(t)-sqrt(2))^2', 'cos(t)^2-2*sqrt(2)*cos(t)+2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '7', '7', 1, 'ATFacForm_int_true.', ''),
        array('FacForm', 'x', '24*(x-1/4)', '24*x-6', 1, 'ATFacForm_default_true.', 'Factors over other fields'),
        array('FacForm', 'x', '(x-sqrt(2))*(x+sqrt(2))', 'x^2-2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', 'x^2-2', 'x^2-2', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(%i*x-2*%i)', '%i*(x-2)', 0, 'ATFacForm_notfactored.', ''),
        array('FacForm', 'x', '%i*(x-2)', '(%i*x-2*%i)', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(x-%i)*(x+%i)', 'x^2+1', 1, 'ATFacForm_true.', ''),
        array('FacForm', 'x', '(x-1)*(x+(1+sqrt(3)*%i)/2)*(x+(1-sqrt(3)*%i)/2)', 'x^3-1', 1, 'ATFacForm_default_true.', ''),

        array('CompSquare', '', '1/0', '0', -1, 'STACKERROR_OPTION.', ''),
        array('CompSquare', 'x', '1/0', '0', -1, 'ATCompSquare_STACKERROR_SAns.', ''),
        array('CompSquare', 'x', '0', '1/0', -1, 'ATCompSquare_STACKERROR_TAns.', ''),
        array('CompSquare', '1/0', '0', '0', -1, 'ATCompSquare_STACKERROR_Opt.', ''),
        array('CompSquare', 'x', '1', '(x-1)^2+1', 0, 'ATCompSquare_SA_not_depend_var.', 'Category errors.'),
        array('CompSquare', 'x', '(t-1)^2+1', '(x-1)^2+1', 0, 'ATCompSquare_SA_not_depend_var.', ''),
        array('CompSquare', 'x', '(x-1)^2+1=0', '(x-1)^2+1', 0, 'ATCompSquare_STACKERROR_LIST.', ''),
        array('CompSquare', 'x', 'sin(x-1)+a-1', '(x-1)^2+1', 0, 'ATCompSquare_false_not_AlgEquiv.', ''),
        array('CompSquare', 'x', '1', '1', 1, 'ATCompSquare_true_trivial.', 'Trivial cases'),
        array('CompSquare', 'x', 'x-a', 'x-a', 1, 'ATCompSquare_true_trivial.', ''),
        array('CompSquare', 'x', 'x^2', 'x^2', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', 'x^2-1', '(x-1)*(x+1)', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '(x-1)^2*k', '(x-1)^2*k', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '(x-1)^2/k', '(x-1)^2/k', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '(x-1)^2+1', '(x-1)^2+1', 1, 'ATCompSquare_true.', 'Normal cases'),
        array('CompSquare', 'x', '(1-x)^2+1', '(x-1)^2+1', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '(X-1)^2+1', '(x-1)^2+1', 0, 'ATCompSquare_SA_not_depend_var.', ''),
        array('CompSquare', 'x', '9*(x-1)^2+1', '(3*x-3)^2+1', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-(x-1)^2', '-(x-1)^2', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-(1-x)^2', '-(x-1)^2', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-(x-1)^2+3', '-(x-1)^2+3', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-(1-x)^2+3', '-(x-1)^2+3', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-4*(x-1)^2+3', '-4*(x-1)^2+3', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-4*(x-1)^2+3', '-(2*x-2)^2+3', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '3-4*(x-1)^2', '-(2*x-2)^2+3', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '(x-1)^2+1', '(x+1)^2+1', 0, 'ATCompSquare_true_not_AlgEquiv.', ''),
        array('CompSquare', 'x', '(x-a^2)^2+1+b', '(x-a^2)^2+1+b', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', 'x^2-2*x+2', '(x-1)^2+1', 0, 'ATCompSquare_false_no_summands.', ''),
        array('CompSquare', 'x', 'x+1', '(x-1)^2+1', 0, 'ATCompSquare_false_not_AlgEquiv.', ''),
        array('CompSquare', 'x', 'a*(x-1)^2+1', 'a*(x-1)^2+1', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'x', '-a*(x-1)^2+1', '1-a*(x-1)^2', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'sin(x)', '(sin(x)-1)^2+1', '(sin(x)-1)^2+1', 1, 'ATCompSquare_true.', 'Not simple variable'),
        array('CompSquare', 'x^2', '(x^2-1)^2+1', '(x^2-1)^2+1', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'y', '(y-1)^2+1', '(y-1)^2+1', 1, 'ATCompSquare_true.', ''),
        array('CompSquare', 'y', '(y+1)^2+1', '(y-1)^2+1', 0, 'ATCompSquare_true_not_AlgEquiv.', ''),
        array('CompSquare', 'sin(x)', '(x-1)^2+1', '(sin(x)-1)^2+1', 0, 'ATCompSquare_SA_not_depend_var.', ''),

        array('PropLogic', '', '1/0', '0', -1, 'ATPropLogic_STACKERROR_SAns.', ''),
        array('PropLogic', '', '0', '1/0', -1, 'ATPropLogic_STACKERROR_TAns.', ''),
        array('PropLogic', '', 'true', 'true', 1, '', ''),
        array('PropLogic', '', 'true', 'false', 0, '', ''),
        array('PropLogic', '', 'A implies B', 'not(A) or B', 1, '', ''),
        array('PropLogic', '', '(a and b and c) xor (a and b) xor (a and c) xor a xor true', '(a implies b) or c', 1, '', ''),

        array('Equiv', '', 'x', '[x^2=4,x=2 or x=-2]', -1, 'ATEquiv_SA_not_list.', ''),
        array('Equiv', '', '[x^2=4,x=2 or x=-2]', 'x',  -1, 'ATEquiv_SB_not_list.', ''),
        array('Equiv', '', '[1/0]', '[x^2=4,x=2 or x=-2]', -1,
            'ATEquiv_STACKERROR_SAns.', ''),
        array('Equiv', '', '[x^2=4,x=2 or x=-2]', '[1/0]',  -1,
            'ATEquiv_STACKERROR_TAns.', ''),
        array('Equiv', '', '[x^2=4,x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1, '[EMPTYCHAR,EQUIVCHAR]', ''),
        array('Equiv', '', '[x^2=4,x=#pm#2,x=2 and x=-2]', '[x^2=4,x=2 or x=-2]', 0, '[EMPTYCHAR,EQUIVCHAR,ANDOR]', ''),
        array('Equiv', '', '[x^2=4,x=2]', '[x^2=4,x=2 or x=-2]', 0, '[EMPTYCHAR,IMPLIEDCHAR]', ''),
        array('Equiv', '[assumepos]', '[x^2=4,x=2]', '[x^2=4,x=2]', 1, '[ASSUMEPOSVARS,EQUIVCHAR]', ''),
        array('Equiv', '', '[x^2=4,x^2-4=0,(x-2)*(x+2)=0,x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1,
            '[EMPTYCHAR,EQUIVCHAR,EQUIVCHAR,EQUIVCHAR]', ''),
        array('Equiv', '', '[x^2=4,x= #pm#2, x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1,
            '[EMPTYCHAR,EQUIVCHAR,EQUIVCHAR]', ''),
        array('Equiv', '', '[x^2-6*x+9=0,x=3]', '[x^2-6*x+9=0,x=3]', 1, '[EMPTYCHAR,SAMEROOTS]', ''),

        array('EquivFirst', '', 'x', '[x^2=4,x=2 or x=-2]', -1, 'ATEquivFirst_SA_not_list.', ''),
        array('EquivFirst', '', '[x^2=4,x=2 or x=-2]', 'x',  -1, 'ATEquivFirst_SB_not_list.', ''),
        array('EquivFirst', '', '[1/0]', '[x^2=4,x=2 or x=-2]', -1,
            'ATEquivFirst_STACKERROR_SAns.', ''),
        array('EquivFirst', '', '[x^2=4,x=2 or x=-2]', '[1/0]',  -1,
            'ATEquivFirst_STACKERROR_TAns.', ''),
        array('EquivFirst', '', '[x^2=4,x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1, '[EMPTYCHAR,EQUIVCHAR]', ''),
        array('EquivFirst', '', '[x^2=9,x=3 or x=-3]', '[x^2=4,x=2 or x=-2]', 0, 'ATEquivFirst_SA_wrong_start', ''),
        array('EquivFirst', '', '[x^2=4,x=2]', '[x^2=4,x=2 or x=-2]', 0, '[EMPTYCHAR,IMPLIEDCHAR]', ''),
        array('EquivFirst', '', '[x^2=4,x^2-4=0,(x-2)*(x+2)=0,x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1,
            '[EMPTYCHAR,EQUIVCHAR,EQUIVCHAR,EQUIVCHAR]', ''),
        array('EquivFirst', '', '[x^2=4,x= #pm#2, x=2 or x=-2]', '[x^2=4,x=2 or x=-2]', 1,
            '[EMPTYCHAR,EQUIVCHAR,EQUIVCHAR]', ''),
        array('EquivFirst', '', '[x^2-6*x+9=0,x=3]', '[x^2-6*x+9=0,x=3]', 1, '[EMPTYCHAR,SAMEROOTS]', ''),

        array('SingleFrac', '', '1/0', '1/n', -1, 'ATSingleFrac_STACKERROR_SAns.', ''),
        array('SingleFrac', '', '0', '1/0', -1, 'ATSingleFrac_STACKERROR_TAns.', ''),
        array('SingleFrac', '', 'x=3', '2', 0, 'ATSingleFrac_SA_not_expression.', ''),
        array('SingleFrac', '', '3', '3', 1, '', ''),
        array('SingleFrac', '', '3', '2', 0, 'ATSingleFrac_ret_exp.', ''),
        array('SingleFrac', '', '1/m', '1/n', 0, 'ATSingleFrac_true. ATSingleFrac_ret_exp.', ''),
        array('SingleFrac', '', '1/n', '1/n', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', 'a+1/2', '(2*a+1)/2', 0, 'ATSingleFrac_part.', ''),
        array('SingleFrac', '', 'a+1/2', '(2*a+1)/2', 0, 'ATSingleFrac_part.', ''),
        array('SingleFrac', '', '4/(x^2+2*x-24)+2/(x^2+4*x-12)', '(6*x-16)/(x^3-28*x+48)', 0, 'ATSingleFrac_part.', ''),
        array('SingleFrac', '', '2*(1/n)', '2/n', 0, 'ATSingleFrac_part.', '2 subtly different answers for the same question'),
        array('SingleFrac', '', '2/n', '2/n', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '2/(n+1)', '1/(n+1)', 0, 'ATSingleFrac_true. ATSingleFrac_ret_exp.', 'Simple Mistakes'),
        array('SingleFrac', '', '(2*n+1)/(n+2)', '1/n', 0, 'ATSingleFrac_true. ATSingleFrac_ret_exp.', ''),
        array('SingleFrac', '', '(2*n)/(n*(n+2))', '(2*n)/(n*(n+3))', 0, 'ATSingleFrac_true. ATSingleFrac_ret_exp.', ''),
        array('SingleFrac', '', '(x-1)/(x^2-1)', '1/(x+1)', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '(1/2)/(3/4)', '2/3', 0, 'ATSingleFrac_div.', 'Fractions within fractions'),
        array('SingleFrac', '', '(x-2)/4/(2/x^2)', '(x-2)*x^2/8', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', '1/(1-1/x)', 'x/(x-1)', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', '(1+1/a)/a', '(1+a)/a^2', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', 'a/(1+1/a)', 'a^2/(1+a)', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', '(1+2*b/a)/c', '(a+2*b)/(a*c)', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', 'c/(1+2*b/a)', 'a*c/(a+2*b)', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', 'a*c/(a+2*b)', 'a*c/(a+2*b)', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '-1/2', '-1/2', 1, 'ATSingleFrac_true.', 'Negative cases'),
        array('SingleFrac', '', '-1/2', '-1/3', 0, 'ATSingleFrac_true. ATSingleFrac_ret_exp.', ''),
        array('SingleFrac', '', '-(1/2)', '-1/2', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '-a/b', '-a/b', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '(-a)/b', '-a/b', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', 'a/(-b)', '-a/b', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '-(a/b)', '-a/b', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '-(1/(n-1))', '1/(1-n)', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', 'a/(-1-1/a)', '-a^2/(1+a)', 0, 'ATSingleFrac_div.', ''),
        array('SingleFrac', '', '((sqrt(5))^3 +6)/15', '((sqrt(5))^3 +6)/15', 1, 'ATSingleFrac_true.', 'Surds in answers'),
        // Use the LowestTerms test for this distinction.
        array('SingleFrac', '', '1/(1-sqrt(2))', '1/(1-sqrt(2))', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '((sqrt(5))^3+6)/15', '((sqrt(5))^3+6)/15', 1, 'ATSingleFrac_true.', ''),
        array('SingleFrac', '', '(5^(3/2)+6)/15', '((sqrt(5))^3+6)/15', 1, 'ATSingleFrac_true.', ''),

        array('PartFrac', '', '1/0', '3*x^2', -1, 'STACKERROR_OPTION.', ''),
        array('PartFrac', 'x', '1/0', '3*x^2', -1, 'ATPartFrac_STACKERROR_SAns.', ''),
        array('PartFrac', '1/0', '0', '0', -1, 'ATPartFrac_STACKERROR_Opt.', ''),
        array('PartFrac', 'x', '0', '1/0', -1, 'ATPartFrac_STACKERROR_TAns.', ''),
        array('PartFrac', 'n', '1/n=0', '1/n', 0, 'ATPartFrac_SA_not_expression.', ''),
        array('PartFrac', 'n', '1/n', '{1/n}', 0, 'ATPartFrac_TA_not_expression.', ''),
        array('PartFrac', 'n', '1/m', '1/n', 0, 'ATPartFrac_diff_variables.', 'Basic tests'),
        array('PartFrac', 'n', '1/n', '1/n', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'n', '1/(n+1)-1/n', '1/(n+1)-1/n', 1, 'ATPartFrac_true.',
            'A simple cases, linear factors in denominator'),
        array('PartFrac', 'n', '1/(n+1)+1/(1-n)', '1/(n+1)-1/(n-1)', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'n', '1/(2*(n-1))-1/(2*(n+1))', '1/((n-1)*(n+1))', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'n', '1/(2*(n+1))-1/(2*(n-1))', '1/((n-1)*(n+1))', 0, 'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '-9/(x-2) + -9/(x+1)', '-9/(x-2) + -9/(x+1)', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '1/(x-1)-(x+1)/(x^2+1)', '2/((x-1)*(x^2+1))', 1, 'ATPartFrac_true.',
            'Irreducible quadratic in denominator'),
        array('PartFrac', 'x', '1/(2*x-2)-(x+1)/(2*(x^2+1))', '1/((x-1)*(x^2+1))', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '1/(2*(x-1))+x/(2*(x^2+1))', '1/((x-1)*(x^2+1))', 0, 'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '3/(x+1) + 3/(x+2)', '3*(2*x+3)/((x+1)*(x+2))', 1, 'ATPartFrac_true.',
            '2 answers to the same question'),
        array('PartFrac', 'x', '3*(1/(x+1) + 1/(x+2))', '3*(2*x+3)/((x+1)*(x+2))', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '3*x*(1/(x+1) + 2/(x+2))', '-12/(x+2)-3/(x+1)+9', 0, 'ATPartFrac_false_degree.',
            'Algebraically equivalent, but numerators of same order than denominator, ie not in partial fraction form.'),
        array('PartFrac', 'x', '(3*x+3)*(1/(x+1) + 2/(x+2))', '9-6/(x+2)', 0, 'ATPartFrac_false_degree.', ''),
        array('PartFrac', 'n', 'n/(2*n-1)-(n+1)/(2*n+1)', '1/(4*n-2)-1/(4*n+2)', 0, 'ATPartFrac_false_degree.', ''),
        array('PartFrac', 'x', '10/(x+3) - 2/(x+2) + x -2', '(x^3 + 3*x^2 + 4*x +2)/((x+2)*(x+3))', 1, 'ATPartFrac_true.',
            'Correct Answer, Numerator > Denominator'),
        array('PartFrac', 'x', '2*x+1/(x+1)+1/(x-1)', '2*x^3/(x^2-1)', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'n', '1/(n*(n-1))', '1/(n*(n-1))', 0, 'ATPartFrac_false_factor.', 'Simple mistakes'),
        array('PartFrac', 'n', '1/(n-1)-1/n^2', '1/((n+1)*n)', 0, 'ATPartFrac_denom_ret.', ''),
        array('PartFrac', 'n', '1/(n-1)-1/n', '1/(n-1)+1/n', 0, 'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '1/(x+1)-1/x', '1/(x-1)+1/x', 0, 'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'n', '1/(n*(n+1))+1/n', '2/n-1/(n+1)', 0, 'ATPartFrac_false_factor.', ''),
        array('PartFrac', 's', '2/(x+1)-1/(x+2)', 's/((s+1)*(s+2))', 0, 'ATPartFrac_diff_variables.',
            'Different Variables'),
        array('PartFrac', 's', 's/((s+1)^2) + s/(s+2) - 1/(s+1)', 's/((s+1)*(s+2))', 0,
            'ATPartFrac_denom_ret.', 'Too many parts in the partial fraction'),
        array('PartFrac', 's', 's/(s+2) - 1/(s+1)', 's/((s+1)*(s+2)*(s+3))', 0,
            'ATPartFrac_denom_ret.', 'Too few parts in the partial fraction'),
        array('PartFrac', 'x', '1/(x+1) + 1/(x+2)', '2/(x+1) + 1/(x+2)', 0,
            'ATPartFrac_ret_expression.', 'Addition and Subtraction errors'),
        array('PartFrac', 'x', '1/(x+1) + 1/(x+2)', '1/(x+1) + 2/(x+2)', 0,
            'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '1/(x+1) + 1/(x+2)', '1/(x+3) + 1/(x+2)', 0,
            'ATPartFrac_ret_expression.', 'Denominator Error'),
        array('PartFrac', 'x', '(2*x+1)/(x^2+1)-2/(x-1)', '(2*x+1)/(x^2+1)-2/(x-1)', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '(-5/(x+3))+(16/(x+3)^2)-(2/(x+2))+4', '(-5/(x+3))+(16/(x+3)^2)-(2/(x+2))+4', 1,
            'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '(3*x^2-5)/((x-4)^2*x)', '(3*x^2-5)/((x-4)^2*x)', 0,
            'ATPartFrac_false_factor.', 'Cubic in the denominator'),
        array('PartFrac', 'x', '-4/(16*x)+53/(16*(x-4))+43/(4*(x-4)^2)', '(3*x^2-5)/((x-4)^2*x)', 0,
            'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '-5/(16*x)+53/(16*(x-4))+43/(4*(x-4)^2)', '(3*x^2-5)/((x-4)^2*x)', 1,
            'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '(5*x+6)/((x+1)*(x+5)^2)', '(5*x+6)/((x+1)*(x+5)^2)', 0,
            'ATPartFrac_false_factor.', ''),
        array('PartFrac', 'x', '-1/(16*(x+5))+19/(4*(x+5)^2)+1/(16*(x+1))', '(5*x+6)/((x+1)*(x+5)^2)', 1,
            'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '5/(x*(x+3)*(5*x-2))', '5/(x*(x+3)*(5*x-2))', 0, 'ATPartFrac_false_factor.', ''),
        array('PartFrac', 'x', '125/(34*(5*x-2))+5/(51*(x+3))-5/(6*x)', '5/(x*(x+3)*(5*x-2))', 1, 'ATPartFrac_true.', ''),
        array('PartFrac', 'x', '(3*x^2-5)/((4*x-4)^2*x)', '(3*x^2-5)/((4*x-4)^2*x)', 0, 'ATPartFrac_false_factor.', ''),
        array('PartFrac', 'x', '-4/(16*x)+1/(2*(x-1))-1/(8*(x-1)^2)', '(3*x^2-5)/((4*x-4)^2*x)', 0,
            'ATPartFrac_ret_expression.', ''),
        array('PartFrac', 'x', '-5/(16*x)+1/(2*(x-1))-1/(8*(x-1)^2)', '(3*x^2-5)/((4*x-4)^2*x)', 1, 'ATPartFrac_true.', ''),

        array('Diff', '', '1/0', '3*x^2', -1, 'STACKERROR_OPTION.', ''),
        array('Diff', '(x', '0', '1/0', -1, 'STACKERROR_OPTION.', ''),
        array('Diff', 'x', '1/0', '3*x^2', -1, 'ATDiff_STACKERROR_SAns.', ''),
        array('Diff', 'x', '0', '1/0', -1, 'ATDiff_STACKERROR_TAns.', ''),
        array('Diff', '1/0', '0', '0', -1, 'ATDiff_STACKERROR_Opt.', ''),
        array('Diff', 'x', '3*x^2', '3*x^2', 1, 'ATDiff_true.', 'Basic tests'),
        array('Diff', 'x', '3*X^2', '3*x^2', 0, 'ATDiff_var_SB_notSA.', ''),
        array('Diff', 'x', 'x^4/4', '3*x^2', 0, 'ATDiff_int.', ''),
        array('Diff', 'x', 'x^4/4+1', '3*x^2', 0, 'ATDiff_int.', ''),
        array('Diff', 'x', 'x^4/4+c', '3*x^2', 0, 'ATDiff_int.', ''),
        array('Diff', 'x', 'y=x^4/4', 'x^4/4', 0, 'ATDiff_SA_not_expression.', ''),
        array('Diff', 'x', 'x^4/4', 'y=x^4/4', 0, '', ''),
        array('Diff', 'x', 'y=x^4/4', 'y=x^4/4', 0, 'ATDiff_SA_not_expression.', ''),
        array('Diff', 'x', '6000*(x-a)^5999', '6000*(x-a)^5999', 1, 'ATDiff_true.', ''),
        array('Diff', 'x', '5999*(x-a)^5999', '6000*(x-a)^5999', 0, '', ''),
        array('Diff', 'x', 'y^2-2*y+1', 'x^2-2*x+1', 0, 'ATDiff_var_SB_notSA.', 'Variable mismatch tests'),
        array('Diff', 'x', 'x^2-2*x+1', 'y^2-2*y+1', 0, 'ATDiff_var_SA_notSB.', ''),
        array('Diff', 'z', 'y^2+2*y+1', 'x^2-2*x+1', 0, 'ATDiff_var_notSASB_SAnceSB.', ''),
        array('Diff', 'y', 'x^4/4', '3*x^2', 0, '', ''),

        array('Int', '', '1/0', '1', -1, 'STACKERROR_OPTION.', ''),
        array('Int', 'x', '1/0', '1', -1, 'ATInt_STACKERROR_SAns.', ''),
        array('Int', 'x', '1', '1/0', -1, 'ATInt_STACKERROR_TAns.', ''),
        array('Int', '1/0', '0', '0', -1, 'ATInt_STACKERROR_Opt.', ''),
        array('Int', '[x,1/0]', '0', '0', -1, 'ATInt_STACKERROR_Opt.', ''),
        array('Int', '[x,NOCONST,1/0]', '0', '0', -1, 'ATInt_STACKERROR_Opt.', ''),
        array('Int', 'x', 'x^3/3', 'x^3/3', 0, 'ATInt_const.', 'Basic tests'),
        array('Int', 'x', 'x^3/3+1', 'x^3/3', 0, 'ATInt_const_int.', ''),
        array('Int', 'x', 'x^3/3+c', 'x^3/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'x^3/3+c+1', 'x^3/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'x^3/3+3*c', 'x^3/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'x^3/3-c', 'x^3/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'x^3/3+c+k', 'x^3/3', 0, 'ATInt_weirdconst.', ''),
        array('Int', 'x', 'x^3/3+c^2', 'x^3/3', 0, 'ATInt_weirdconst.', ''),
        array('Int', 'x', 'x^3/3*c', 'x^3/3', 0, 'ATInt_generic.', ''),
        array('Int', 'x', 'X^3/3+c', 'x^3/3', 0, 'ATInt_generic. ATInt_var_SB_notSA.', ''),
        array('Int', 'x', 'sin(2*x)', 'x^3/3', 0, 'ATInt_generic.', ''),
        array('Int', 'x', 'x^2/2-2*x+2+c', '(x-2)^2/2', 1, 'ATInt_true.', ''),
        array('Int', 't', '(t-1)^5/5+c', '(t-1)^5/5', 1, 'ATInt_true.', ''),
        array('Int', 'v', '(v-1)^5/5+c', '(v-1)^5/5', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'cos(2*x)/2+1+c', 'cos(2*x)/2', 1, 'ATInt_true.', ''),
        array('Int', 'x', '(x-a)^6001/6001+c', '(x-a)^6001/6001', 1, 'ATInt_true.', ''),
        array('Int', 'x', '(x-a)^6001/6001', '(x-a)^6001/6001', 0, 'ATInt_const.', ''),
        array('Int', 'x', '6000*(x-a)^5999', '(x-a)^6001/6001', 0, 'ATInt_diff.', ''),
        array('Int', 'x', '4*%e^(4*x)/(%e^(4*x)+1)', 'log(%e^(4*x)+1)+c', 0, 'ATInt_generic.', ''),
        array('Int', 'x', 'x^3/3+c', 'x^3/3+c', 1, 'ATInt_true.', 'The teacher adds a constant'),
        array('Int', 'x', 'x^2/2-2*x+2+c', '(x-2)^2/2+k', 1, 'ATInt_true.', ''),
        array('Int', '[x,NOCONST]', 'x^3/3', 'x^3/3', 1, 'ATInt_const_condone.',
            'The teacher condones lack of constant, or numerical constant'),
        array('Int', '[x,NOCONST]', 'x^3/3+c', 'x^3/3', 1, 'ATInt_true.', ''),
        array('Int', '[x,NOCONST]', 'x^2/2-2*x+2', '(x-2)^2/2+k', 1, 'ATInt_const_condone.', ''),
        array('Int', '[x,NOCONST]', 'x^3/3+1', 'x^3/3', 1, 'ATInt_const_int_condone.', ''),
        array('Int', '[x,NOCONST]', 'x^3/3+c^2', 'x^3/3', 0, 'ATInt_weirdconst.', ''),
        array('Int', 'x', 'exp(x)+c', 'exp(x)', 1, 'ATInt_true.', 'Special case'),
        array('Int', 'x', 'exp(x)', 'exp(x)', 0, 'ATInt_const.', ''),
        array('Int', '[x,NOCONST]', 'exp(x)', 'exp(x)', 1, 'ATInt_const_condone.', ''),
        array('Int', 'x', '2*x', 'x^3/3', 0, 'ATInt_diff.', 'Student differentiates by mistake'),
        array('Int', 'x', '2*x+c', 'x^3/3', 0, 'ATInt_diff.', ''),
        array('Int', 'x', 'ln(x)', 'ln(x)', 0, 'ATInt_const.', 'Sloppy logs (teacher ignores abs(x) )'),
        array('Int', '[x,NOCONST]', 'ln(x)', 'ln(x)', 1, 'ATInt_const_condone.', ''),
        array('Int', 'x', 'ln(x)+c', 'ln(x)+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(k*x)', 'ln(x)+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(x)', 'ln(abs(x))+c', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', 'Fussy logs (teacher uses abs(x) )'),
        array('Int', 'x', 'ln(x)+c', 'ln(abs(x))+c', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', '[x, NOCONST]', 'ln(x)', 'ln(abs(x))+c', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'ln(abs(x))', 'ln(abs(x))+c', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'ln(abs(x))+c', 'ln(abs(x))+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(k*x)', 'ln(abs(x))+c', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'ln(k*abs(x))', 'ln(abs(x))+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(abs(k*x))', 'ln(abs(x))+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(x)', 'ln(k*abs(x))', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', 'Teacher uses ln(k*abs(x))'),
        array('Int', 'x', 'ln(x)+c', 'ln(k*abs(x))', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'ln(abs(x))', 'ln(k*abs(x))', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'ln(abs(x))+c', 'ln(k*abs(x))', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(k*x)', 'ln(k*abs(x))', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'ln(k*abs(x))', 'ln(k*abs(x))', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'ln(x)+ln(a)', 'ln(k*abs(x+a))', 0, 'ATInt_generic. ATInt_logabs.', 'Other logs'),
        array('Int', 'x', 'log(x)^2-2*log(c)*log(x)+k', 'ln(c/x)^2', 0, 'ATInt_EqFormalDiff.', ''),
        array('Int', 'x', 'log(x)^2-2*log(c)*log(x)+k', 'ln(abs(c/x))^2', 0, 'ATInt_generic.', ''),
        array('Int', 'x', 'c-(log(2)-log(x))^2/2', '-1/2*log(2/x)^2', 1, 'ATInt_true_equiv.', ''),
        // In these examples there are two logarihtms.  The student should be *consistent*
        // in their use, or not, of absolute value.
        array('Int', 'x', 'log(abs(x-3))+log(abs(x+3))', 'log(abs(x-3))+log(abs(x+3))', 0, 'ATInt_const.', 'Two logs'),
        array('Int', 'x', 'log(abs(x-3))+log(abs(x+3))+c', 'log(abs(x-3))+log(abs(x+3))', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'log(abs(x-3))+log(abs(x+3))', 'log(x-3)+log(x+3)', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'log(abs(x-3))+log(abs(x+3))+c', 'log(x-3)+log(x+3)', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'log(x-3)+log(x+3)', 'log(x-3)+log(x+3)', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'log(x-3)+log(x+3)+c', 'log(x-3)+log(x+3)', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'log(x-3)+log(x+3)', 'log(abs(x-3))+log(abs(x+3))', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'log(x-3)+log(x+3)+c', 'log(abs(x-3))+log(abs(x+3))', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 'x', 'log(abs((x-3)*(x+3)))+c', 'log(abs(x-3))+log(abs(x+3))', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', 'log(abs((x^2-9)))+c', 'log(abs(x-3))+log(abs(x+3))', 0, 'ATInt_EqFormalDiff.', ''),
        // This comes from the integral of x^3+2*x^2-3*x-2)/(x^2-4).
        array('Int', 'x', '2*log(abs(x-2))-log(abs(x+2))+(x^2+4*x)/2',
            '-log(abs(x+2))+2*log(abs(x-2))+(x^2+4*x)/2+c', 0, 'ATInt_const.', ''),
        array('Int', 'x', '-log(abs(x+2))+2*log(abs(x-2))+(x^2+4*x)/2+c',
            '-log(abs(x+2))+2*log(abs(x-2))+(x^2+4*x)/2+c', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', '-log(abs(x+2))+2*log(abs(x-2))+(x^2+4*x)/2+c',
            '-log((x+2))+2*log((x-2))+(x^2+4*x)/2', 1, 'ATInt_true_equiv.', ''),
        // Inconsistent cases. (Teacher doesn't use abs).
        array('Int', 'x', 'log(abs(x-3))+log((x+3))+c', 'log(x-3)+log(x+3)', 0,
            'ATInt_true_equiv. ATInt_logabs_inconsistent.', 'Inconsistent log(abs())'),
        array('Int', 'v', 'log((v-3))+log(abs(v+3))+c', 'log(v-3)+log(v+3)', 0,
            'ATInt_true_equiv. ATInt_logabs_inconsistent.', ''),
        array('Int', 'x', 'log((x-3))+log(abs(x+3))', 'log(x-3)+log(x+3)', 0,
            'ATInt_const. ATInt_logabs_inconsistent.', ''),
        array('Int', 'x', '2*log((x-2))-log(abs(x+2))+(x^2+4*x)/2',
            '-log(abs(x+2))+2*log(abs(x-2))+(x^2+4*x)/2', 0, 'ATInt_EqFormalDiff. ATInt_logabs. ATInt_logabs_inconsistent.', ''),
        array('Int', 't', '2*(sqrt(t)-5)-10*log((sqrt(t)-5))+c',
            '2*(sqrt(t)-5)-10*log((sqrt(t)-5))+c', 1, 'ATInt_true_equiv.', 'Significant integration constant differences'),
        array('Int', 't', '2*(sqrt(t))-10*log((sqrt(t)-5))+c',
            '2*(sqrt(t)-5)-10*log((sqrt(t)-5))+c', 1, 'ATInt_true_differentconst.', ''),
        array('Int', 't', '2*(sqrt(t)-5)-10*log((sqrt(t)-5))+c',
            '2*(sqrt(t)-5)-10*log(abs(sqrt(t)-5))+c', 0, 'ATInt_EqFormalDiff. ATInt_logabs.', ''),
        array('Int', 't', '2*(sqrt(t))-10*log(abs(sqrt(t)-5))+c',
            '2*(sqrt(t)-5)-10*log(abs(sqrt(t)-5))+c', 1, 'ATInt_true_differentconst.', ''),
        array('Int', 'x', '2*sin(x)*cos(x)', 'sin(2*x)+c', 0, 'ATInt_const.', 'Trig'),
        array('Int', 'x', '2*sin(x)*cos(x)+k', 'sin(2*x)+c', 1, 'ATInt_true.', ''),
        array('Int', 'x', '-2*cos(3*x)/3-3*cos(2*x)/2', '-2*cos(3*x)/3-3*cos(2*x)/2+c', 0, 'ATInt_const.', ''),
        array('Int', 'x', '-2*cos(3*x)/3-3*cos(2*x)/2+1', '-2*cos(3*x)/3-3*cos(2*x)/2+c', 0, 'ATInt_const_int.', ''),
        array('Int', 'x', '-2*cos(3*x)/3-3*cos(2*x)/2+c', '-2*cos(3*x)/3-3*cos(2*x)/2+c', 1, 'ATInt_true.', ''),
        array('Int', 't', '(tan(2*t)-2*t)/2',
            '-(t*sin(4*t)^2-sin(4*t)+t*cos(4*t)^2+2*t*cos(4*t)+t)/(sin(4*t)^2+cos(4*t)^2+2*cos(4*t)+1)', 0, 'ATInt_const.', ''),
        array('Int', 't', '(tan(2*t)-2*t)/2+1',
            '-(t*sin(4*t)^2-sin(4*t)+t*cos(4*t)^2+2*t*cos(4*t)+t)/(sin(4*t)^2+cos(4*t)^2+2*cos(4*t)+1)', 0, 'ATInt_const_int.', ''),
        array('Int', 't', '(tan(2*t)-2*t)/2+c',
            '-(t*sin(4*t)^2-sin(4*t)+t*cos(4*t)^2+2*t*cos(4*t)+t)/(sin(4*t)^2+cos(4*t)^2+2*cos(4*t)+1)', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'tan(x)-x+c', 'tan(x)-x', 1, 'ATInt_true.', ''),
        array('Int', 'x', '((5*%e^7*x-%e^7)*%e^(5*x))', '((5*%e^7*x-%e^7)*%e^(5*x))/25+c', 0,
            'ATInt_generic.', 'Note the difference in feedback here, generated by the options.'),
        array('Int', '[x,x*%e^(5*x+7)]', '((5*%e^7*x-%e^7)*%e^(5*x))', '((5*%e^7*x-%e^7)*%e^(5*x))/25+c', 0, 'ATInt_generic.', ''),
        // Various forms of inverse hyperbolic forms of the integrals.  Consider int(1/(x^2-a^2),x).
        array('Int', 'x', 'log(x-3)/6-log(x+3)/6+c', 'log(x-3)/6-log(x+3)/6', 1, 'ATInt_true_equiv.',
            'Inverse hyperbolic integrals'),
        array('Int', 'x', 'asinh(x)', 'ln(x+sqrt(x^2+1))', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'asinh(x)+c', 'ln(x+sqrt(x^2+1))', 1, 'ATInt_true.', ''),
        array('Int', 'x', '-acoth(x/3)/3', 'log(x-3)/6-log(x+3)/6', 0, 'ATInt_const.', ''),
        array('Int', '[x, NOCONST]', '-acoth(x/3)/3', 'log(x-3)/6-log(x+3)/6', 1, 'ATInt_true.', ''),
        array('Int', 'x', '-acoth(x/3)/3+c', 'log(x-3)/6-log(x+3)/6', 1, 'ATInt_true.', ''),
        array('Int', 'x', '-acoth(x/3)/3+c', 'log(abs(x-3))/6-log(abs(x+3))/6', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'log(x-a)/(2*a)-log(x+a)/(2*a)+c', 'log(x-a)/(2*a)-log(x+a)/(2*a)', 1, 'ATInt_true_equiv.', ''),
        array('Int', 'x', '-acoth(x/a)/a+c', 'log(x-a)/(2*a)-log(x+a)/(2*a)', 1, 'ATInt_true.', ''),
        array('Int', 'x', '-acoth(x/a)/a+c', 'log(abs(x-a))/(2*a)-log(abs(x+a))/(2*a)', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'log(x-a)/(2*a)-log(x+a)/(2*a)+c', 'log(abs(x-a))/(2*a)-log(abs(x+a))/(2*a)', 0,
            'ATInt_EqFormalDiff. ATInt_logabs.', ''),

        array('Int', 'x', 'log(x-3)/6-log(x+3)/6+c', '-acoth(x/3)/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'log(abs(x-3))/6-log(abs(x+3))/6+c', '-acoth(x/3)/3', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'log(x-3)/6-log(x+3)/6', '-acoth(x/3)/3', 0, 'ATInt_const.', ''),
        // Non-trivial example from JHD, July 2017.
        array('Int', 'x', 'atan(2*x-3)+c', 'atan(2*x-3)', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'atan((x-2)/(x-1))+c', 'atan(2*x-3)', 1, 'ATInt_true.', ''),
        array('Int', 'x', 'atan((x-2)/(x-1))', 'atan(2*x-3)', 0, 'ATInt_const.', ''),
        array('Int', 'x', 'atan((x-1)/(x-2))', 'atan(2*x-3)', 0, 'ATInt_generic.', ''),
        // These onse currently fail for mathematical reasons.
        array('Int', 'x', '2/3*sqrt(3)*(atan(sin(x)/(sqrt(3)*(cos(x)+1)))-(atan(sin(x)/(cos(x)+1))))+x/sqrt(3)',
            '2*atan(sin(x)/(sqrt(3)*(cos(x)+1)))/sqrt(3)', -3, 'ATInt_const.', 'Stoutemyer (currently fails)'),

        array('GT', '', '1/0', '1', -1, 'ATGT_STACKERROR_SAns.', ''),
        array('GT', '', '1', '1/0', -1, 'ATGT_STACKERROR_TAns.', ''),
        array('GT', '', '1', '1', 0, 'ATGT_false.', ''),
        array('GT', '', '2', '1', 1, 'ATGT_true.', ''),
        array('GT', '', '1', '2.1', 0, 'ATGT_false.', ''),
        array('GT', '', 'pi', '3', 1, 'ATGT_true.', ''),
        array('GT', '', 'pi+2', '5', 1, 'ATGT_true.', ''),
        array('GT', '', '-inf', '0', 0, 'Not number', 'Infinity'),
        array('GT', '', 'inf', '0', 0, 'Not number', ''),

        array('GTE', '', '1/0', '1', -1, 'ATGTE_STACKERROR_SAns.', ''),
        array('GTE', '', '1', '1/0', -1, 'ATGTE_STACKERROR_TAns.', ''),
        array('GTE', '', '1', '1', 1, 'ATGTE_true.', ''),
        array('GTE', '', '2', '1', 1, 'ATGTE_true.', ''),
        array('GTE', '', '1', '2.1', 0, 'ATGTE_false.', ''),
        array('GTE', '', 'pi', '3', 1, 'ATGTE_true.', ''),
        array('GTE', '', 'pi+2', '5', 1, 'ATGTE_true.', ''),

        array('NumRelative', '', '1/0', '0', -1, 'ATNumRelative_STACKERROR_SAns.', 'Basic tests'),
        array('NumRelative', '', '0', '1/0', -1, 'ATNumRelative_STACKERROR_TAns.', ''),
        array('NumRelative', '1/0', '0', '0', -1, 'ATNumRelative_STACKERROR_Opt.', ''),
        array('NumRelative', '', '0', '(x', -1, 'ATNumRelativeTEST_FAILED-Empty TA.', ''),
        array('NumRelative', 'x', '1.5', '1.5', -1, 'ATNumerical_STACKERROR_tol.', ''),
        // Invalid options should be caught at edit time.  If they get this far it will be taken as the default 0.05.
        array('NumRelative', '(x', '1', '0', 0, '', ''),
        array('NumRelative', '', 'x=1.5', '1.5', 0, 'ATNumerical_FAILED.', ''),
        array('NumRelative', '', '1.1', '1', 0, '', 'No option, so 5%'),
        array('NumRelative', '', '1.05', '1', 1, '', ''),
        array('NumRelative', '', '0.95', '1', 1, '', ''),
        array('NumRelative', '', '0.949', '1', 0, '', ''),
        array('NumRelative', '', '1.05e33', '1e33', 1, '', ''),
        array('NumRelative', '', '1.06e33', '1e33', 0, '', ''),
        array('NumRelative', '', '0.95e33', '1e33', 1, '', ''),
        array('NumRelative', '', '0.949e33', '1e33', 0, '', ''),
        array('NumRelative', '', '1.05e-33', '1e-33', 1, '', ''),
        array('NumRelative', '', '1.06e-33', '1e-33', 0, '', ''),
        array('NumRelative', '', '0.95e-33', '1e-33', 1, '', ''),
        array('NumRelative', '', '0.949e-33', '1e-33', 0, '', ''),
        array('NumRelative', '0.1', '1.05', '1', 1, '', 'Options passed'),
        array('NumRelative', '0.1', '1.05', '3', 0, '', ''),
        array('NumRelative', '0.001', '3.14', 'pi', 1, '', ''),
        array('NumRelative', '', 'inf', '0', 0, 'ATNumerical_FAILED.', 'Infinity'),
        array('NumRelative', '', '1', '[1,2]', 0, 'ATNumerical_SA_not_list.', 'Lists'),
        array('NumRelative', '', '[1,2]', '[1,2,3]', 0, 'ATNumerical_wronglen.', ''),
        array('NumRelative', '', '[1,2]', '[1,2]', 1, '', ''),
        array('NumRelative', '', '[3.141,1.414]', '[pi,sqrt(2)]', 1, '', ''),
        array('NumRelative', '0.01', '[3,1.414]', '[pi,sqrt(2)]', 0, 'ATNumerical_wrongentries SA/TA=[3.0].', ''),
        array('NumRelative', '0.01', '[3,1.414]', '{pi,sqrt(2)}', 0, 'ATNumerical_SA_not_set.', ''),
        array('NumRelative', '0.01', '{1.414,3.1}', '{significantfigures(pi,6),sqrt(2)}', 0,
            'ATNumerical_wrongentries: TA/SA=[3.14159], SA/TA=[3.1].', ''),
        array('NumRelative', '0.1', '{1.414,3.1}', '{pi,sqrt(2)}', 1, '', ''),

        array('NumAbsolute', '', '1/0', '0', -1, 'ATNumAbsolute_STACKERROR_SAns.', 'Basic tests'),
        array('NumAbsolute', '', '0', '1/0', -1, 'ATNumAbsolute_STACKERROR_TAns.', ''),
        array('NumAbsolute', '1/0', '0', '0', -1, 'ATNumAbsolute_STACKERROR_Opt.', ''),
        array('NumAbsolute', '', '0', '(x', -1, 'ATNumAbsoluteTEST_FAILED-Empty TA.', ''),
        // Invalid options should be caught at edit time.  If they get this far it will be taken as the default 0.05.
        array('NumAbsolute', '(x', '1', '0', 0, '', ''),
        array('NumAbsolute', '', '1.1', '1', 0, '', 'No option, so 5%'),
        array('NumAbsolute', '', '1.05', '1', 1, '', ''),
        array('NumAbsolute', '0.1', '1.05', '1', 1, '', 'Options passed'),
        array('NumAbsolute', '0.1', '1.05', '3', 0, '', ''),
        array('NumAbsolute', '0.001', '3.14', 'pi', 0, '', ''),
        array('NumAbsolute', '0.0001', '1.41e-2', '1.41e-2', 1, '', ''),
        array('NumAbsolute', '0.0001', '0.0141', '1.41e-2', 1, '', ''),
        array('NumAbsolute', '0.0001', '0.00141', '0.00141', 1, '', ''),
        array('NumAbsolute', '0.0001', '0.00141', '1.41*10^-3', 1, '', ''),
        array('NumAbsolute', '0.0001', '1.41*10^-3', '1.41*10^-3', 1, '', ''),
        array('NumAbsolute', '0.01', '[3.141,1.414]', '[pi,sqrt(2)]', 1, '', ''),
        array('NumAbsolute', '0.01', '[3,1.414]', '[pi,sqrt(2)]', 0, 'ATNumerical_wrongentries SA/TA=[3.0].', ''),
        array('NumAbsolute', '0.01', '[3,1.414]', '{pi,sqrt(2)}', 0, 'ATNumerical_SA_not_set.', ''),
        array('NumAbsolute', '0.01', '{1.414,3.1}', '{significantfigures(pi,6),sqrt(2)}', 0,
            'ATNumerical_wrongentries: TA/SA=[3.14159], SA/TA=[3.1].', ''),
        array('NumAbsolute', '0.1', '{1,1.414,3.1,2}', '{1,2,pi,sqrt(2)}', 1, '', ''),

        array('NumSigFigs', '', '3.141', '3.1415927', -1, 'STACKERROR_OPTION.', 'Basic tests'),
        array('NumSigFigs', '3', '1/0', '3', -1, 'ATNumSigFigs_STACKERROR_SAns.', ''),
        array('NumSigFigs', '3', '0', '1/0', -1, 'ATNumSigFigs_STACKERROR_TAns.', ''),
        array('NumSigFigs', '1/0', '0', '0', -1, 'ATNumSigFigs_STACKERROR_Opt.', ''),
        array('NumSigFigs', '(', '0', '1', -1, 'STACKERROR_OPTION.', ''),
        array('NumSigFigs', '1', '(', '1', -1, 'ATNumSigFigsTEST_FAILED-Empty SA.', ''),
        array('NumSigFigs', 'pi', '1', '3', -1, 'ATNumSigFigs_STACKERROR_not_integer.', ''),
        array('NumSigFigs', '[3,x]', '1', '3', -1, 'ATNumSigFigs_STACKERROR_not_integer.', ''),
        array('NumSigFigs', '[1,2,3]', '1', '3', -1, 'ATNumSigFigs_STACKERROR_list_wrong_length.', ''),
        array('NumSigFigs', '', '1', '3', -1, 'STACKERROR_OPTION.', ''),
        array('NumSigFigs', '4', 'pi', 'pi', 0, 'ATNumSigFigs_NotDecimal.', ''),
        array('NumSigFigs', '2', '0', '0', 0, 'ATNumSigFigs_WrongDigits.', 'Edge cases'),
        array('NumSigFigs', '1', '0', '0', 1, '', ''),
        array('NumSigFigs', '1', '0.0', '0', 1, '', ''),
        array('NumSigFigs', '2', '0.0', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '2', '0', '0.0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '2', '0.0', '0.0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '2', '0.00', '0.00', 1, '', ''),
        array('NumSigFigs', '2', '5.4e21', '5.3e21', 0, 'ATNumSigFigs_Inaccurate.', 'Large numbers'),
        array('NumSigFigs', '2', '5.3e21', '5.3e21', 1, '', ''),
        array('NumSigFigs', '2', '5.3e22', '5.3e22', 1, '', ''),
        array('NumSigFigs', '2', '5.3e20', '5.3e22', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '9', '6.02214086e23', '6.02214086e23', 1, '', ''),
        array('NumSigFigs', '9', '6.0221409e23', '6.02214086e23', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '9', '6.02214087e23', '6.02214086e23', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '9', '6.02214085e23', '6.02214086e23', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '8', '5.3910632e-44', '5.3910632e-44', 1, '', ''),
        array('NumSigFigs', '8', '5.391063e-44', '5.3910632e-44', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '8', '5.3910631e-44', '5.3910632e-44', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '8', '5.3910633e-44', '5.3910632e-44', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '9', '1.61622938e-35', '1.61622938e-35', 1, '', ''),
        array('NumSigFigs', '9', '1.6162294e-35', '1.61622938e-35', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '9', '1.61622939e-35', '1.61622938e-35', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '9', '1.61622937e-35', '1.61622938e-35', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '5', '1.2345e82', '1.2345e82', 1, '', ''),
        array('NumSigFigs', '5', '1.2346e82', '1.2345e82', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '5', '1.2344e82', '1.2345e82', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '1', '1.234', '4', 0,
                'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', 'No trailing zeros.'),
        array('NumSigFigs', '3', '3.141', '3.1415927', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '4', '3.141', '3.1415927', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '3.146', '3.1415927', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '3.147', '3.1415927', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '4', '3.142', '3.1415927', 1, '', ''),
        array('NumSigFigs', '4', '3.142', 'pi', 1, '', ''),
        array('NumSigFigs', '4', '3141', '3.1415927', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '3', '0.00123', '0.001234567', 1, '', ''),
        array('NumSigFigs', '3', '1.23e-3', '0.001234567', 1, '', ''),
        array('NumSigFigs', '3', '138*10^-3', '138*10^-3', 1, '', ''),
        array('NumSigFigs', '3', '-138*10^-3', '-138*10^-3', 1, '', ''),
        array('NumSigFigs', '3', '138*10^-3', '-138*10^-3', 0, 'ATNumSigFigs_WrongSign.', ''),
        array('NumSigFigs', '3', '1.38*10^-1', '138*10^-3', 1, '', ''),
        array('NumSigFigs', '3', '1.24e-3', '0.001234567', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '1.235e-3', '0.001234567', 1, '', ''),
        array('NumSigFigs', '2', '1000', '999', 1, 'ATNumSigFigs_WithinRange.', ''),
        array('NumSigFigs', '2', '1E3', '999', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '1', '-100', '-149', 1, '', ''),
        array('NumSigFigs', '1', '-0.05', '-0.0499', 1, '', ''),
        array('NumSigFigs', '1', '-(0.05)', '-0.0499', 1, '', ''),
        array('NumSigFigs', '3', '1170', '1174.34', 1, '', ''),
        array('NumSigFigs', '3', '61300', '61250', 1, '', ''),
        array('NumSigFigs', '4', '0.1667', '0.1667', 1, '', 'Previous tricky case'),
        array('NumSigFigs', '4', '0.1666', '0.1667', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '0.1663', '0.1667', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '0.1662', '0.1667', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '4', '0.166', '0.1667', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '4', '0.16667', '0.1667', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '4', '-3.141', '-3.1415927', 0, 'ATNumSigFigs_Inaccurate.', 'Negative numbers'),
        array('NumSigFigs', '3', '-3.141', '-3.1415927', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '4', '-3.141', '-3.1415927', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '4', '-3.142', '-3.1415927', 1, '', ''),
        array('NumSigFigs', '4', '3.142', '-3.1415927', 0, 'ATNumSigFigs_WrongSign.', ''),
        array('NumSigFigs', '4', '-3.142', '3.1415927', 0, 'ATNumSigFigs_WrongSign.', ''),
        array('NumSigFigs', '4', '-3.149', '3.1415927', 0, 'ATNumSigFigs_WrongSign. ATNumSigFigs_VeryInaccurate.', ''),
        // Note that 75701719/35227192=2.148956947803276, so this tests rounding in the teacher's answer.
        array('NumSigFigs', '3', '2.15', '75701719/35227192', 1, '', ''),
        // Maxima's round() command uses Bankers' rounding, but STACK does not.
        // We actually round the teacher's answer to the specified number of SFs.
        array('NumSigFigs', '3', '0.0499', '0.04985', 1, '', 'Round teacher answer'),
        array('NumSigFigs', '3', '0.0498', '0.04985', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '3', '0.0498', '0.04975', 1, '', ''),
        array('NumSigFigs', '3', '0.0497', '0.04975', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '3', '0.0499', '0.0498', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '3', '1.5', '1.500', 0, 'ATNumSigFigs_WrongDigits.',
            'Final zeros after the decimal are significant.'),
        array('NumSigFigs', '3', '1.50', '1.500', 1, '', ''),
        array('NumSigFigs', '3', '1.500', '1.500', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '3', '245.0', '245', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // The test DOES NOT recognize too few significant figures being entered.
        array('NumSigFigs', '3', '180', '178.35', 0, 'ATNumSigFigs_WithinRange. ATNumSigFigs_Inaccurate.', 'Too few digits'),
        array('NumSigFigs', '3', '33', '33.1558', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', ''),
        // 0.0010 has exactly 2 significant digits.
        // With mixed options [n,m] we check for n significant figures, and that the student's answer
        // matches the teacher's answer to m of them.
        array('NumSigFigs', '[4,3]', '3.142', '3.1415927', 1, '', 'Mixed options'),
        // In this test case there are 4 sig figs, only 3 of which are accurate. Should generate feedback.
        array('NumSigFigs', '[4,3]', '3.143', '3.1415927', 1, '', ''),
        // In this test case there are 4 sig figs, only 2 of which are accurate. Should generate feedback.
        array('NumSigFigs', '[4,3]', '3.150', '3.1415927', 0, 'ATNumSigFigs_Inaccurate.', ''),
        // In this test case there are 4 sig figs, only 1 of which are accurate. Should not generate feedback.
        array('NumSigFigs', '[4,3]', '3.211', '3.1415927', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        // In this test case there are 5 sig figs, which is the wrong number.
        array('NumSigFigs', '[4,3]', '3.1416', '3.1415927', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[4,3]', '0.1666', '0.1667', 1, '', ''),
        // Range of sigfigs of 180 contains 3, accurate to 1!
        array('NumSigFigs', '[3,1]', '180', '178.35', 1, 'ATNumSigFigs_WithinRange.', ''),
        array('NumSigFigs', '[3,1]', '33', '33.1558', 0, 'ATNumSigFigs_WrongDigits.', ''), // Too few sigfigs.
        array('NumSigFigs', '[3,1]', '1.500', '1.5', 0, 'ATNumSigFigs_WrongDigits.', ''), // Too many sigfigs.
        array('NumSigFigs', '[3,1]', '245.0', '245', 0, 'ATNumSigFigs_WrongDigits.', ''), // Too many sigfigs.
        // This example has rounding.
        array('NumSigFigs', '[6,6]', '12345.7', '12345.654321', 1, '', ''),
        array('NumSigFigs', '[6,3]', '12345.7', '12345.654321', 1, '', ''),
        array('NumSigFigs', '[6,3]', '12300.0', '12345.654321', 1, '', ''),
        array('NumSigFigs', '[6,3]', '12400.0', '12345.654321', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[6,3]', '13500.0', '12345.654321', 0, 'ATNumSigFigs_VeryInaccurate.', ''),
        array('NumSigFigs', '[6,2]', '12000.0', '12345.654321', 1, '', ''),
        array('NumSigFigs', '[6,2]', '13000.0', '12345.654321', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[6,2]', '11000.0', '12345.654321', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[1,0]', '0.0010', '0', 0, 'ATNumSigFigs_WrongDigits.', 'Zero option and trailing zeros'),
        array('NumSigFigs', '[2,0]', '0.0010', '0', 1, '', ''),
        array('NumSigFigs', '[3,0]', '0.0010', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 0.001 has exactly 1 significant digits.
        array('NumSigFigs', '[1,0]', '0.001', '0', 1, '', ''),
        array('NumSigFigs', '[2,0]', '0.001', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 0.00100 has exactly 3 significant digits.
        array('NumSigFigs', '[2,0]', '0.00100', 'null', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[3,0]', '0.00100', 'null', 1, '', ''),
        array('NumSigFigs', '[4,0]', '0.00100', 'null', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 5.00 has exactly 3 significant digits.
        array('NumSigFigs', '[2,0]', '5.00', 'null', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[3,0]', '5.00', 'null', 1, '', ''),
        array('NumSigFigs', '[4,0]', '5.00', 'null', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 100 has at least 1 and maybe even 3 significant digits.
        array('NumSigFigs', '[1,0]', '100', '0', 1, '', ''),
        array('NumSigFigs', '[2,0]', '100', '0', 1, 'ATNumSigFigs_WithinRange.', ''),
        array('NumSigFigs', '[3,0]', '100', '0', 1, 'ATNumSigFigs_WithinRange.', ''),
        array('NumSigFigs', '[4,0]', '100', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 10.0 has exactly 3 significant digits.
        array('NumSigFigs', '[2,0]', '10.0', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[3,0]', '10.0', '0', 1, '', ''),
        array('NumSigFigs', '[4,0]', '10.0', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 0 has exactly 1 significant digits.
        array('NumSigFigs', '[1,0]', '0', '0', 1, '', ''),
        array('NumSigFigs', '[2,0]', '0', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // 0.00 has at exactly 2 significant digits.
        array('NumSigFigs', '[1,0]', '0.00', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[2,0]', '0.00', '0', 1, '', ''),
        array('NumSigFigs', '[3,0]', '0.00', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[4,0]', '0.00', '0', 0, 'ATNumSigFigs_WrongDigits.', ''),
        // Condone too many significant figures.
        array('NumSigFigs', '[4,-1]', '8.250', '8.250', 1, '', 'Condone too many sfs.'),
        array('NumSigFigs', '[4,-1]', '8.25', '8.250', 0, 'ATNumSigFigs_WrongDigits.', ''),
        array('NumSigFigs', '[4,-1]', '8.250000', '8.250', 1, '', ''),
        array('NumSigFigs', '[4,-1]', '8.250434', '8.250', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '82.4', '82', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '82.5', '82', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[2,-1]', '83', '82', 0, 'ATNumSigFigs_Inaccurate.', ''),
        // Check where teacher uses a rational number 1/7 = 0.142857142857...
        array('NumSigFigs', '[4,-1]', '0.1430', '1/7', 0, 'ATNumSigFigs_Inaccurate.', '1/7 = 0.142857142857...'),
        array('NumSigFigs', '[4,-1]', '0.1429', '1/7', 1, '', ''),
        array('NumSigFigs', '[4,-1]', '0.1428', '1/7', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[4,-1]', '0.143', '1/7', 0, 'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate.', ''),
        // Too many sig figs, which is condoned.
        array('NumSigFigs', '[4,-1]', '0.14284', '1/7', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[4,-1]', '0.14285', '1/7', 1, '', ''),
        array('NumSigFigs', '[4,-1]', '0.14286', '1/7', 1, '', ''),
        array('NumSigFigs', '[4,-1]', '0.14291', '1/7', 1, '', ''),
        array('NumSigFigs', '[4,-1]', '0.14294', '1/7', 1, '', ''),
        // Incorrectly rounded means to 4 s.f. this is too large.
        array('NumSigFigs', '[4,-1]', '0.14295', '1/7', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '[2,-1]', '0.142', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.14290907676', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.143', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.1433333', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.144', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.145', '1/7', 1, '', ''),
        array('NumSigFigs', '[2,-1]', '0.146', '1/7', 0, 'ATNumSigFigs_Inaccurate.', ''),
        // Teacher does not give a float.
        array('NumSigFigs', '4', '1.279', 'ev(lg(19),lg=logbasesimp)', 1, '', 'Logarithms, numbers and surds'),
        array('NumSigFigs', '3', '3.14', 'pi', 1, '', ''),
        array('NumSigFigs', '3', '3.15', 'pi', 0, 'ATNumSigFigs_Inaccurate.', ''),
        array('NumSigFigs', '6', '1.73205', 'sqrt(3)', 1, '', ''),
        array('NumSigFigs', '2', 'matrix([0.33,1],[1,1])', 'matrix([0.333,1],[1,1])', -1, 'ATNumSigFigs_NotDecimal.',
                'No support for matrices!'),
        array('NumSigFigs', '2', '3.1415', 'matrix([0.333,1],[1,1])', -1, 'TEST_FAILED', ''),
        array('NumSigFigs', '3', '1.50', 'dispsf(1.500,3)', 1, '', 'Teacher uses dispsf'),

        array('NumDecPlaces', '2', '1/0', '3', -1, 'ATNumDecPlaces_STACKERROR_SAns.', 'Basic tests'),
        array('NumDecPlaces', '2', '0.1', '1/0', -1, 'ATNumDecPlaces_STACKERROR_TAns.', ''),
        array('NumDecPlaces', '1/0', '0.1', '0', -1, 'ATNumDecPlaces_STACKERROR_Opt.', ''),
        array('NumDecPlaces', 'x', '0.1', '1', -1, 'ATNumDecPlaces_OptNotInt.', ''),
        array('NumDecPlaces', '-1', '0.1', '1', -1, 'ATNumDecPlaces_OptNotInt.', ''),
        array('NumDecPlaces', '0', '0.1', '1', -1, 'ATNumDecPlaces_OptNotInt.', ''),
        array('NumDecPlaces', '(', '0.1', '1', -1, 'STACKERROR_OPTION.', ''),
        array('NumDecPlaces', '1', '(', '1', -1, 'ATNumDecPlacesTEST_FAILED-Empty SA.', ''),
        array('NumDecPlaces', '2', 'x', '3.143', 0,
            'ATNumDecPlaces_SA_Not_num.', 'Student\'s answer not a floating point number'),
        array('NumDecPlaces', '3', 'pi', '3.000', 0, 'ATNumDecPlaces_SA_Not_num.', ''),
        array('NumDecPlaces', '2', '3.14', '3.143', 1,
            'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', 'Right number of places'),
        array('NumDecPlaces', '2', '3.14', '3.14', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '3', '3.140', '3.140', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '4', '3141.5972', '3141.5972', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '2', '4.14', '3.14', 0, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Not_equiv.', ''),
        array('NumDecPlaces', '4', '3.1416', 'pi', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '1', '-7.3', '-7.3', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '1', '3.14', '3.143', 0,
            'ATNumDecPlaces_Wrong_DPs. ATNumDecPlaces_Equiv.', 'Wrong number of places'),
        array('NumDecPlaces', '1', '3.14', '3.143', 0, 'ATNumDecPlaces_Wrong_DPs. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '3', '3.14', '3.140', 0, 'ATNumDecPlaces_Wrong_DPs. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '4', '7.000', '7', 0, 'ATNumDecPlaces_Wrong_DPs. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '4', '7.0000', '7', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.', ''),
        array('NumDecPlaces', '3', '8.0000', '7', 0, 'ATNumDecPlaces_Wrong_DPs. ATNumDecPlaces_Not_equiv.',
            'Both wrong DPs and inaccurate.'),
        array('NumDecPlaces', '3', '4.000', '3.99999', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.',
            'Teacher needs to round their answer.'),
        array('NumDecPlaces', '2', '0.10', 'displaydp(0.1,2)', 1, 'ATNumDecPlaces_Correct. ATNumDecPlaces_Equiv.',
            'Teacher uses displaydp'),

        array('NumDecPlacesWrong', '2', '1/0', '3', -1, 'ATNumDecPlacesWrong_STACKERROR_SAns.', 'Basic tests'),
        array('NumDecPlacesWrong', '2', '0.1', '1/0', -1, 'ATNumDecPlacesWrong_STACKERROR_TAns.', ''),
        array('NumDecPlacesWrong', '1/0', '0.1', '0', -1, 'ATNumDecPlacesWrong_STACKERROR_Opt.', ''),
        array('NumDecPlacesWrong', 'x', '0.1', '0', -1, 'ATNumDecPlacesWrong_OptNotInt.', ''),
        array('NumDecPlacesWrong', '4', 'x^2', '1234', 0, 'ATNumDecPlacesWrong_SA_Not_num.', ''),
        array('NumDecPlacesWrong', '4', '1234.5', 'x^2', 0, 'ATNumDecPlacesWrong_Tans_Not_Num.', ''),
        array('NumDecPlacesWrong', '4', '3.141', '31.41', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '4', '3.141', '31.14', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '4', 'pi', '31.14', 0, 'ATNumDecPlacesWrong_SA_Not_num.', ''),
        array('NumDecPlacesWrong', '4', '0.1234', '1234', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '4', '0.1235', '1234', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '4', '0.0001234', '1234', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '4', '0.0001235', '1234', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '3', '0.1233', '1234', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '0.1243', '1234', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '3', '0.1230', '1239', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '0.1240', '1239', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '3', '1230', '1239', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '2230', '1239', 0, 'ATNumDecPlacesWrong_Wrong.', ''),
        array('NumDecPlacesWrong', '3', '0.100', '1.00', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '0.1000', '1.00', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '0.1001', '1.001', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '4', '0.100', '1.0', 1, 'ATNumDecPlacesWrong_Correct.', 'Condone lack of trailing zeros'),
        array('NumDecPlacesWrong', '4', '1', '1.00', 1, 'ATNumDecPlacesWrong_Correct.', ''),
        array('NumDecPlacesWrong', '3', '0.101', 'displaydp(101,3)', 1, 'ATNumDecPlacesWrong_Correct.',
                'Teacher uses displaydp'),

        array('SigFigsStrict', '', '3.141', 'null', -1, 'STACKERROR_OPTION.', 'Basic tests'),
        array('SigFigsStrict', 'x^2', '3.141', 'null', -1, 'STACKERROR_OPTION.', ''),
        array('SigFigsStrict', '-2', '3.141', 'null', -1, 'STACKERROR_OPTION.', ''),
        array('SigFigsStrict', '0', '3.141', 'null', -1, 'STACKERROR_OPTION.', ''),
        // 0.0010 has exactly 2 significant digits.
        array('SigFigsStrict', '1', '0.0010', 'null', 0, '', ''),
        array('SigFigsStrict', '2', '0.0010', 'null', 1, '', ''),
        array('SigFigsStrict', '3', '0.0010', 'null', 0, '', ''),
        // 0.00100 has exactly 3 significant digits.
        array('SigFigsStrict', '2', '0.00100', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '0.00100', 'null', 1, '', ''),
        array('SigFigsStrict', '4', '0.00100', 'null', 0, '', ''),
        // 0.001 has exactly 1 significant digits.
        array('SigFigsStrict', '1', '0.001', 'null', 1, '', ''),
        array('SigFigsStrict', '2', '0.001', 'null', 0, '', ''),
        // 100 has exactly 1 significant digit.
        array('SigFigsStrict', '1', '100', 'null', 1, '', ''),
        array('SigFigsStrict', '2', '100', 'null', 0, 'ATSigFigsStrict_WithinRange.', ''),
        array('SigFigsStrict', '3', '100', 'null', 0, 'ATSigFigsStrict_WithinRange.', ''),
        array('SigFigsStrict', '4', '100', 'null', 0, '', ''),
        // 100. has exactly 3 significant digit.
        array('SigFigsStrict', '1', '100.', 'null', 0, '', ''),
        array('SigFigsStrict', '2', '100.', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '100.', 'null', 1, '', ''),
        array('SigFigsStrict', '4', '100.', 'null', 0, '', ''),
        // 123. has exactly 3 significant digit.
        array('SigFigsStrict', '1', '123.', 'null', 0, '', ''),
        array('SigFigsStrict', '2', '123.', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '123.', 'null', 1, '', ''),
        array('SigFigsStrict', '4', '123.', 'null', 0, '', ''),
        // 1.00e2 has exactly 3 significant digit2.
        array('SigFigsStrict', '1', '1.00e2', 'null', 0, '', ''),
        array('SigFigsStrict', '2', '1.00e2', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '1.00e2', 'null', 1, '', ''),
        array('SigFigsStrict', '4', '1.00e2', 'null', 0, '', ''),
        // 10.0 has exactly 3 significant digits.
        array('SigFigsStrict', '2', '10.0', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '10.0', 'null', 1, '', ''),
        array('SigFigsStrict', '4', '10.0', 'null', 0, '', ''),
        // 0 has exactly 1 significant digits.
        array('SigFigsStrict', '1', '0', 'null', 1, '', ''),
        array('SigFigsStrict', '2', '0', 'null', 0, '', ''),
        // 0.0 has exactly 1 significant digits.
        array('SigFigsStrict', '1', '0.0', 'null', 1, '', ''),
        array('SigFigsStrict', '2', '0.0', 'null', 0, '', ''),
        // Accept that .0 has exactly 1 significant digits.
        array('SigFigsStrict', '1', '.0', 'null', 1, '', ''),
        array('SigFigsStrict', '2', '.0', 'null', 0, '', ''),
        // Accept that .001030 has exactly 4 significant digits.
        array('SigFigsStrict', '4', '.001030', 'null', 1, '', ''),
        // 0.00 has exactly 2 significant digits.
        array('SigFigsStrict', '1', '0.00', 'null', 0, '', ''),
        array('SigFigsStrict', '2', '0.00', 'null', 1, '', ''),
        array('SigFigsStrict', '3', '0.00', 'null', 0, '', ''),
        // Mix of notations.
        array('SigFigsStrict', '1', '25.00e1', 'null', 0, '', ''),
        array('SigFigsStrict', '3', '25.00e1', 'null', 0, '', ''),
        array('SigFigsStrict', '4', '25.00e1', 'null', 1, '', ''),
        array('SigFigsStrict', '5', '25.00e1', 'null', 0, '', ''),
        // Too many zeros.
        array('SigFigsStrict', '3', '15.1', '15.1', 1, '', ''),
        array('SigFigsStrict', '3', '15.10', '15.1', 0, '', ''),
        array('SigFigsStrict', '3', '15.100', '15.1', 0, '', ''),
        array('SigFigsStrict', '3', '9.81*m/s^2', 'null', 1, '', 'Units are ignored'),

        array('Units', '2', '1/0', '1', -1, 'ATUnits_STACKERROR_SAns.', ''),
        array('Units', '2', '1', '1/0', -1, 'ATUnits_STACKERROR_TAns.', ''),
        array('Units', '1/0', '1', '1', -1, 'ATUnits_STACKERROR_Opt.', ''),
        array('Units', '2', 'x-1)^2', '(x-1)^2', -1, 'ATUnitsTEST_FAILED-Empty SA.', ''),
        array('Units', '[3,x]', '12.3*m*s^(-1)', '3*m', -1, 'ATNumSigFigs_STACKERROR_not_integer.', ''),
        array('Units', '[1,2,3]', '3*m*s^(-1)', '3*m', -1,
            'ATNumSigFigs_STACKERROR_list_wrong_length.', ''),
        array('Units', '3', '12.3*m*s^(-1)', '{12.3*m*s^(-1)}', -1, 'ATUnits_TA_not_expression.', ''),
        array('Units', '3', 'x=12.3*m*s^(-1)', '12.3*m*s^(-1)', 0, 'ATUnits_SA_not_expression.', ''),
        array('Units', '3', '12.3', '12.3*m', 0, 'ATUnits_SA_no_units.', 'Missing units'),
        array('Units', '3', '12', '12.3*m', 0, 'ATUnits_SA_no_units.', ''),
        array('Units', '3', '1/2', '12.3*m', 0, 'ATUnits_SA_no_units.', ''),
        array('Units', '3', 'e^(1/2)', '12.3*m', 0, 'ATUnits_SA_no_units.', ''),
        array('Units', '3', '9.81*m', '12.3', -1, 'ATUnits_SB_no_units.', ''),
        array('Units', '3', 'm/s', '12.3*m/s', 0, 'ATUnits_SA_only_units.', 'Only units'),
        array('Units', '3', 'm', '12.3*m/s', 0, 'ATUnits_SA_only_units.', ''),
        array('Units', '3', '9.81+m/s', '9.81*m/s', 0, 'ATUnits_SA_bad_units.', 'Bad units'),
        array('Units', '3', '12.3*m/s', '12.3*m/s', 1, 'ATUnits_units_match.', 'Basic tests'),
        array('Units', '3', '12.4*m/s', '12.3*m/s', 0, 'ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '[3,2]', '12.4*m/s', '12.3*m/s', 1, 'ATUnits_units_match.', ''),
        array('Units', '[3,2]', '12.45*m/s', '12.3*m/s', 0, 'ATNumSigFigs_WrongDigits. ATUnits_units_match.', ''),
        array('Units', '[3,2]', '13.45*m/s', '12.3*m/s', 0,
            'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '[3,2]', '7.54E-5*(s*M)^-1', '5.625E-5*s^-1', 0,
            'ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', ''),
        array('Units', '[3,2]', '7.54E-5*(s*M)^-1', 'stackunits(5.625E-5,1/s)', 0,
            'ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', ''),
        array('Units', '3', '12*m/s', '12.3*m/s', 0,
            'ATNumSigFigs_WrongDigits. ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m/s^2', '-9.81*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '-9.82*m/s^2', '-9.815*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m/s^2', '-9.815*m/s^2', 0, 'ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m*s^(-2)', '-9.81*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '-9.82*m/s^2', '-9.81*m/s^2', 0, 'ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m*s^(-2)', '-9.81*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m/s/s', '-9.81*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '-9.81*m/s', '-9.81*m/s^2', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('Units', '3', '-9.81*m/s', '-9.81*m/s^2', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        // In the following test case the student uses brackets which are not needed, but broke the test.
        array('Units', '3', '(-9.81)*m/s^2', '-9.81*m/s^2', 1, 'ATUnits_units_match.', ''),
        array('Units', '3', '520*amu', '520*amu', 1, 'ATNumSigFigs_WithinRange. ATUnits_units_match.', ''),
        array('Units', '3', '520*amu', '521*amu', 0, 'ATNumSigFigs_WithinRange. ATNumSigFigs_Inaccurate. ATUnits_units_match.', ''),
        array('Units', '3', '(-9.81)', '-9.81*m/s^2', 0, 'ATUnits_SA_no_units.', 'Missing units'),
        array('Units', '3', '9.81*m/s', '-9.81*m/s^2', 0, 'ATNumSigFigs_WrongSign. ATUnits_incompatible_units.', ''),
        array('Units', '3', '8.81*m/s', '-9.81*m/s^2', 0,
            'ATNumSigFigs_WrongSign. ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', ''),
        array('Units', '3', '8.1*m/s', '-9.81*m/s^2', 0,
            'ATNumSigFigs_WrongDigits. ATNumSigFigs_WrongSign. ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', ''),
        array('Units', '3', 'm/4', '0.25*m', 0, 'ATNumSigFigs_WrongDigits. ATUnits_units_match.', ''),
        array('Units', '3', 'pi*s', '3.14*s', 0, 'ATNumSigFigs_WrongDigits. ATUnits_units_match.', 'Student is too exact'),
        array('Units', '3', 'sqrt(2)*m', '1.41*m', 0, 'ATNumSigFigs_WrongDigits. ATUnits_units_match.', ''),
        array('Units', '2', '25*g', '0.025*kg', 1, 'ATUnits_compatible_units kg.', 'Different units'),
        array('Units', '2', '26*g', '0.025*kg', 0, 'ATNumSigFigs_Inaccurate. ATUnits_compatible_units kg.', ''),
        array('Units', '2', '100*g', '10*kg', 0,
            'ATNumSigFigs_WithinRange. ATNumSigFigs_VeryInaccurate. ATUnits_compatible_units kg.', ''),
        array('Units', '2', '0.025*g', '0.025*kg', 0, 'ATUnits_compatible_units kg. ATUnits_correct_numerical.', ''),
        array('Units', '2', '1000*m', '1*km', 1, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m.', ''),
        array('Units', '1', '1*Mg/10^6', '1*N*s^2/(km)', 1, 'ATUnits_compatible_units kg.', ''),
        array('Units', '1', '1*Mg/10^6', '1*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.', ''),
        array('Units', '3', '3.14*Mg/10^6', '%pi*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.', ''),
        array('Units', '3', '3.141*Mg/10^6', '%pi*kN*ns/(mm*Hz)', 0, 'ATNumSigFigs_WrongDigits. ATUnits_compatible_units kg.', ''),
        array('Units', '3', '4.141*Mg/10^6', '%pi*kN*ns/(mm*Hz)', 0,
           'ATNumSigFigs_WrongDigits. ATNumSigFigs_VeryInaccurate. ATUnits_compatible_units kg.', ''),
        array('Units', '2', '400*cc', '0.4*l', 1, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('Units', '2', '400*cm^3', '0.4*l', 1, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('Units', '2', '400*ml', '0.4*l', 1, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('Units', '2', '18*kJ', '18000.0*J', 1, 'ATUnits_compatible_units (kg*m^2)/s^2.', ''),
        array('Units', '2', '18.1*kJ', '18000.0*J', 0, 'ATNumSigFigs_WrongDigits. ATUnits_compatible_units (kg*m^2)/s^2.', ''),
        array('Units', '2', '2.0*hh', '720000*s', 1, 'ATUnits_compatible_units s.', ''),
        array('Units', '3', '723*kVA', '0.723*MVA', 1, 'ATUnits_compatible_units (kg*m^2)/s^3.', ''),
        array('Units', '1', '0*m/s', '0*m/s', 1, 'ATUnits_units_match.', 'Edge case'),
        array('Units', '1', '0.0*m/s', '0*m/s', 1, 'ATUnits_units_match.', ''),
        array('Units', '1', '0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('Units', '2', '0.00*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('Units', '1', '0.0*km/s', '0.0*m/s', 1, 'ATUnits_compatible_units m/s.', ''),
        array('Units', '1', '0.0*m', '0.0*m/s', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('Units', '1', '0.0', '0.0*m/s', 0, 'ATUnits_SA_no_units.', ''),
        array('Units', '1', '7*in', '7*in', 1, 'ATUnits_units_match.', 'Imperial'),
        array('Units', '1', '6*in', '0.5*ft', 1, 'ATUnits_compatible_units in.', ''),
        array('Units', '4', '2640*ft', '0.5*mi', 1, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units in.', ''),
        array('Units', '4', '2650*ft', '0.5*mi', 0,
            'ATNumSigFigs_WithinRange. ATNumSigFigs_VeryInaccurate. ATUnits_compatible_units in.', ''),
        array('Units', '4', '142.8*C', '415.9*K', -3, 'ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', 'TODO'),
        // Atomic mass unit: numbers out of range.
        array('Units', '3', '520*mamu', '520*mamu', -3, 'ATUnits_SB_no_units.', ''),

        array('UnitsStrict', '2', '25*g', '0.025*kg', 0, 'ATUnits_compatible_units kg.', 'Differences from the Units test only'),
        array('UnitsStrict', '1', '1*Mg/10^6', '1*N*s^2/(km)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsStrict', '1', '1*Mg/10^6', '1*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsStrict', '3', '3.14*Mg/10^6', '%pi*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsStrict', '2', '400*cc', '0.4*l', 0, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('UnitsStrict', '2', '400*cm^3', '0.4*l', 0, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('UnitsStrict', '2', '400*ml', '0.4*l', 0, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('UnitsStrict', '2', '400*mL', '0.4*l', 0, 'ATNumSigFigs_WithinRange. ATUnits_compatible_units m^3.', ''),
        array('UnitsStrict', '4', '142.8*C', '415.9*K', 0, 'ATNumSigFigs_VeryInaccurate. ATUnits_incompatible_units.', ''),
        array('UnitsStrict', '3', '-9.81*m/s/s', '-9.81*m/s^2', 1, 'ATUnits_units_match.', 'We are not *that* strict!'),
        array('UnitsStrict', '1', '0*m/s', '0*m/s', 1, 'ATUnits_units_match.', 'Edge case'),
        array('UnitsStrict', '1', '0.0*m/s', '0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '1', '0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '1', '0.0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '1', '0.0*km/s', '0.0*m/s', 0, 'ATUnits_compatible_units m/s.', ''),
        array('UnitsStrict', '1', '0.0*m', '0.0*m/s', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('UnitsStrict', '1', '0.0', '0.0*m/s', 0, 'ATUnits_SA_no_units.', ''),
        array('UnitsStrict', '[3,2]', '2.33e-15*kg', '2.33e-15*kg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '7.03e-3*ng', '7.03e-3*ng', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '2.35e-6*ug', '2.35e-6*ug', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.83e-10*cg', '9.83e-10*cg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.73e-21*Gg', '9.73e-21*Gg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '7.19e-15*kg', '7.19e-15*kg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '8.12e-12*g', '8.12e-12*g', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.34e-12*g', '9.34e-12*g', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '1.07e-21*Gg', '1.07e-21*Gg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '1.91e-10*cg', '1.91e-10*cg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '5.67e-18*Mg', '5.67e-18*Mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '2.04e-9*mg', '2.04e-9*mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '6.75e-6*ug', '6.75e-6*ug', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '6.58e-6*ug', '6.58e-6*ug', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '3.58e-9*mg', '3.58e-9*mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.99e-15*kg', '9.99e-15*kg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.8e-9*mg', '9.8e-9*mg', 0,
            'ATNumSigFigs_WrongDigits. ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.80e-9*mg', '9.8e-9*mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.83e-9*mg', '9.8e-9*mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '[3,2]', '9.78e-9*mg', '9.8e-9*mg', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '36*Kj/mol', '36*Kj/mol', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '-36*Kj/mol', '-36*Kj/mol', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '(-36)*Kj/mol', '-36*Kj/mol', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '(-36*Kj)/mol', '-36*Kj/mol', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '-(36*Kj)/mol', '-36*Kj/mol', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrict', '2', '-(36.2*Kj)/mol', '-36.3*Kj/mol', 0,
            'ATNumSigFigs_WrongDigits. ATUnits_units_match.', ''),

        array('UnitsRelative', '0.01', '12.3*m/s', '12.3*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsRelative', '0.01', '12*m/s', '12.3*m/s', 0, 'ATUnits_units_match.', ''),
        array('UnitsRelative', '0.15', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.', ''),
        array('UnitsRelative', '0.05', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsRelative', '0.01', '0*m/s', '0*m/s', 1, 'ATUnits_units_match.', 'Edge case'),
        array('UnitsRelative', '0.01', '0.0*m/s', '0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsRelative', '0.01', '0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsRelative', '0.01', '0.0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsRelative', '0.01', '0.0*km/s', '0.0*m/s', 1, 'ATUnits_compatible_units m/s.', ''),
        array('UnitsRelative', '0.01', '0.0*m', '0.0*m/s', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('UnitsRelative', '0.01', '0.0', '0.0*m/s', 0, 'ATUnits_SA_no_units.', ''),

        array('UnitsStrictRelative', '0.01', '12.3*m/s', '12.3*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrictRelative', '0.01', '12*m/s', '12.3*m/s', 0, 'ATUnits_units_match.', ''),
        array('UnitsStrictRelative', '0.15', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsStrictRelative', '0.05', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsStrictRelative', '0.01', '0*m/s', '0*m/s', 1, 'ATUnits_units_match.', 'Edge case'),
        array('UnitsStrictRelative', '0.01', '0.0*m/s', '0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrictRelative', '0.01', '0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrictRelative', '0.01', '0.0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsStrictRelative', '0.01', '0.0*km/s', '0.0*m/s', 0, 'ATUnits_compatible_units m/s.', ''),
        array('UnitsStrictRelative', '0.01', '0.0*m', '0.0*m/s', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('UnitsStrictRelative', '0.01', '0.0', '0.0*m/s', 0, 'ATUnits_SA_no_units.', ''),
        array('UnitsStrictRelative', '0.01', '0*J', '0.0*J', 1, 'ATUnits_units_match.', ''),

        array('UnitsAbsolute', '5*J', '-123000*J', '-123*kJ', 0, 'ATUnits_SO_wrong_units.', ''),
        array('UnitsAbsolute', '0.01', '12.3*m/s', '12.3*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '12*m/s', '12.3*m/s', 0, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.15', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.', ''),
        array('UnitsAbsolute', '0.1', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.',
                'The following illustrates that we convert to base units to compare.'),
        array('UnitsAbsolute', '0.09', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsAbsolute', '5*kJ', '-123000*J', '-123*kJ', 1, 'ATUnits_compatible_units (kg*m^2)/s^2.', 'Units in the options'),
        array('UnitsAbsolute', '5*kJ', '-123006*J', '-123*kJ', 1, 'ATUnits_compatible_units (kg*m^2)/s^2.', ''),
        array('UnitsAbsolute', '5*kJ', '-129006*J', '-123*kJ', 0, 'ATUnits_compatible_units (kg*m^2)/s^2.', ''),
        array('UnitsAbsolute', '0.1*kN*ns/(mm*Hz)', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 1, 'ATUnits_compatible_units kg.', ''),
        array('UnitsAbsolute', '0.09*kN*ns/(mm*Hz)', '1.1*Mg/10^6', '1.2*kN*ns/(mm*Hz)', 0, 'ATUnits_compatible_units kg.', ''),
        array('UnitsAbsolute', '0.01', '0*m/s', '0*m/s', 1, 'ATUnits_units_match.', 'Edge case'),
        array('UnitsAbsolute', '0.01', '0.0*m/s', '0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '0.0*m/s', '0.0*m/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '0.0*km/s', '0.0*m/s', 1, 'ATUnits_compatible_units m/s.', ''),
        array('UnitsAbsolute', '0.01', '0.0*m', '0.0*m/s', 0, 'ATUnits_incompatible_units. ATUnits_correct_numerical.', ''),
        array('UnitsAbsolute', '0.01', '0.0', '0.0*m/s', 0, 'ATUnits_SA_no_units.', ''),
        // The teacher's answer here is likely to be the result of internal simplification.
        array('UnitsAbsolute', '0.01', '1.0*m/s', 'm/s', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '15/pi*kN/mm^2', '15/pi*kN/mm^2', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '(15*kN)/(pi*mm^2)', '(15*kN)/(pi*mm^2)', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '(15/pi)*(kN/mm^2)', '(15/pi)*(kN/mm^2)', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '(600*N)/(%pi*mm^2)', '(600*N)/(%pi*mm^2)', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '(600/pi)*kN/m^2', '(600/pi)*kN/m^2', 1, 'ATUnits_units_match.', ''),
        array('UnitsAbsolute', '0.01', '(600/pi)*kN/mm^2', '(600/pi)*kN/mm^2', 1, 'ATUnits_units_match.', ''),

        array('String', '', 'Hello', 'hello', 0, '', ''),
        array('String', '', 'hello', 'hello', 1, '', ''),
        array('String', '', 'hello', 'heloo', 0, '', ''),

        array('StringSloppy', '', 'hello', 'Hello', 1, '', ''),
        // This is a change.  The ast does not have acess to the raw string typed by the student.
        array('StringSloppy', '', 'hel lo', 'Hello', 0, 'ATStringSloppy_STACKERROR_SAns.', ''),
        array('StringSloppy', '', 'hel lo', 'Hel*lo', 0, 'ATStringSloppy_STACKERROR_SAns.', ''),
        array('StringSloppy', '', 'hello', 'heloo', 0, '', ''),

        array('SRegExp', '', '1/0', '"hello"', -1, 'ATSRegExp_STACKERROR_SAns.', ''),
        array('SRegExp', '', '"1/0"', '1/0', -1, 'ATSRegExp_STACKERROR_TAns.', ''),
        array('SRegExp', '', 'Hello', 'hello', -1, 'ATSRegExp_SB_not_string.', ''),
        array('SRegExp', '', 'Hello', '"hello"', -1, 'ATSRegExp_SA_not_string.', ''),
        array('SRegExp', '', '"aaaaabbb"', '"(aaa)*b"', 1, 'ATSRegExp: ["aaab","aaa"].', ''),
        array('SRegExp', '', '"aab"', '"(aaa)*b"', 1, 'ATSRegExp: ["b",false].', ''),
        array('SRegExp', '', '"aaac"', '"(aaa)*b"', 0, '', ''),
        array('SRegExp', '', '"aab"', '"^[aA]*b$"', 1, 'ATSRegExp: ["aab"].',
            'Anchor pattern to the start and the end of the string'),
        array('SRegExp', '', '"aab"', '"^(aaa)*b$"', 0, '', ''),
        array('SRegExp', '', '"aAb"', '"^[aA]*b$"', 1, 'ATSRegExp: ["aAb"].', ''),
        array('SRegExp', '', '" aAb"', '"^[aA]*b$"', 0, '', ''),
        array('SRegExp', '', '"caAb"', '"(?i:a*b)"', 1, 'ATSRegExp: ["aAb"].', 'Case insensitive'),
        array('SRegExp', '', '"Alice went to the market"', '"(Alice|Bob) went to the (bank|market)"', 1,
            'ATSRegExp: ["Alice went to the market","Alice","market"].', 'Options'),
        array('SRegExp', '', '"Malice went to the shop"', '"(Alice|Bob) went to the (bank|market)"', 0, '', ''),
        array('SRegExp', '', '"Alice   went  to      the market"', '"(Alice|Bob)\\\\s+went\\\\s+to\\\\s+the\\\\s+(bank|market)"',
            1, 'ATSRegExp: ["Alice   went  to      the market","Alice","market"].', 'Whitespace, ' .
            'note test rendering issue, the test string has additional spaces and tabs as does the result'),
        array('SRegExp', '', '"Alice   went  to      themarket"',
            '"(Alice|Bob)\\\\s+went\\\\s+to\\\\s+the\\\\s+(bank|market)"', 0, '', ''),
        array('SRegExp', '', '"x^2.2"', '"x\\\\^2\\\\.2"', 1, 'ATSRegExp: ["x^2.2"].',
            'Escaping patterns, note the function that does it'),
        array('SRegExp', '', '"x^2+sin(x)"', 'sconcat(string_to_regex("sin(x)"),"$")', 1, 'ATSRegExp: ["sin(x)"].', ''),
        array('SRegExp', '', '"sin(x)+x^2"', 'sconcat(string_to_regex("sin(x)"),"$")', 0, '', ''),

        array('LowestTerms', '', '1/0', '0', -1, 'ATLowestTerms_STACKERROR_SAns.', ''),
        array('LowestTerms', '', '0.5', '0', 1, '', 'Mix of floats and rational numbers'),
        array('LowestTerms', '', '0.33', '0', 1, '', ''),
        array('LowestTerms', '', '2/4', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '-1/3', '0', 1, '', 'Negative numbers'),
        array('LowestTerms', '', '1/-3', '0', 1, '', ''),
        array('LowestTerms', '', '-2/4', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '2/-4', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '-1/-3', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '-2/-4', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'x+1/3', '0', 1, '', 'Polynomials'),
        array('LowestTerms', '', 'x+2/6', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '2*x/4+2/6', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', '2/4*x+2/6', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'x-1/-4', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'cos(x)', '0', 1, '', 'Trig functions'),
        array('LowestTerms', '', 'cos(3/6*x)', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'matrix([1,2/4],[2,3])', '0', 0, 'ATLowestTerms_entries.', 'Matrices'),
        array('LowestTerms', '', 'x=1/2', '0', 1, '', 'Equations'),
        array('LowestTerms', '', '3/9=x', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'x^2/x', '0', 1, '', 'Use predicate lowesttermsp'),
        array('LowestTerms', '', '(2*x)/(4*t)', '0', 1, '', ''),
        array('LowestTerms', '', '(2/4)*(x^2/t)', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'x^(2/4)', '0', 0, 'ATLowestTerms_entries.', ''),
        array('LowestTerms', '', 'sqrt(3)/3', 'sqrt(3)/3', 1, '', 'Need to rationalize demoninator'),
        array('LowestTerms', '', '1/sqrt(3)', 'sqrt(3)/3', 0, 'ATLowestTerms_not_rat.', ''),
        array('LowestTerms', '', '1/(1-sqrt(2))', '1/(1-sqrt(2))', 0, 'ATLowestTerms_not_rat.', ''),
        array('LowestTerms', '', '1/(1+i)', '(1-i)/2', 0, 'ATLowestTerms_not_rat.', ''),
        array('LowestTerms', '', '1+2/sqrt(3)', '(2*sqrt(3)+3)/3', 0, 'ATLowestTerms_not_rat.', ''),
        array('LowestTerms', '', '1/(1+1/root(3,2))', 'sqrt(3)/(sqrt(3)+1)', 0, 'ATLowestTerms_not_rat.', ''),
        array('LowestTerms', '', '1/(1+1/root(2,3))', '1/(1+1/root(2,3))', 0, 'ATLowestTerms_not_rat.', '')

    );

    public static function get_raw_test_data() {
        $equiv = new stack_equiv_test_data();
        return array_merge(self::$rawdata, $equiv->get_answertestfixtures());
    }

    public static function get_available_tests() {
        $availabletests = array();
        foreach (self::$rawdata as $test) {
            $availabletests[$test[self::NAME]] = $test[self::NAME];
        }
        return $availabletests;
    }

    public static function test_from_raw($data) {

        $test = new stdClass();
        $test->name          = $data[self::NAME];
        $test->studentanswer = $data[self::SANS];
        $test->teacheranswer = $data[self::TANS];
        $test->expectedscore = $data[self::SCORE];
        $test->options       = $data[self::OPTIONS];
        $test->ansnote       = $data[self::ANSNOTE];
        $test->notes         = $data[self::NOTES];
        return $test;
    }

    public static function get_all() {
        $tests = array();
        $rawdata = self::get_raw_test_data();
        foreach ($rawdata as $data) {
            $tests[] = self::test_from_raw($data);
        }
        return $tests;
    }

    public static function get_tests_for($anstest) {
        $tests = array();
        $rawdata = self::get_raw_test_data();

        foreach ($rawdata as $data) {
            if ($data[self::NAME] == $anstest) {
                $tests[] = self::test_from_raw($data);
            }
        }
        return $tests;
    }

    public static function run_test($test) {
        $sans = stack_ast_container::make_from_teacher_source($test->studentanswer, '', new stack_cas_security());
        $tans = stack_ast_container::make_from_teacher_source($test->teacheranswer, '', new stack_cas_security());
        $topt = stack_ast_container::make_from_teacher_source($test->options, '', new stack_cas_security());

        $anst = new stack_ans_test_controller($test->name, $sans, $tans, $topt, new stack_options());

        // The false clause is useful for developers to track down which test case is breaking Maxima.
        if (true) {
            $result   = $anst->do_test(); // This actually executes the answer test in the CAS.
            $errors   = $anst->get_at_errors();
            $rawmark  = $anst->get_at_mark();
            $feedback = $anst->get_at_feedback();
            $ansnote  = $anst->get_at_answernote();
        } else {
            $feedback = 'AT'.$test->name.'('.$test->studentanswer.','.$test->teacheranswer.');';
            $result   = true; // This actually executes the answer test in the CAS.
            $errors   = '';
            $rawmark  = 0;
            $ansnote  = '';
        }

        $trace = $anst->get_trace(false);
        $anomalynote = array();
        $passed = true;
        if ($test->expectedscore >= 0) {
            if ($rawmark !== $test->expectedscore) {
                $passed = false;
                $anomalynote[] = '[SCORE]';
            }
        } else {
            // We expect the test to fail.
            switch ($test->expectedscore) {
                case -1:
                    if ($errors !== '') {
                        $passed = true;
                    } else {
                        $passed = false;
                        $anomalynote[] = '[Expected test to fail!]';
                    }
                    break;

                case -2:
                    if ($rawmark) {
                        $passed = true;
                    } else {
                        $passed = false;
                        $anomalynote[] = '[Expected maths failure: got 0, expected 1.]';
                    }
                    break;

                case -3:
                    if ($rawmark) {
                        $passed = false;
                        $anomalynote[] = '[Expected maths failure: got 1, expected 0.]';
                    } else {
                        $passed = true;
                    }
                    break;

                default:
                    $passed = false;
                    $anomalynote[] = '[General failure.]';
            }
        }

        if (!($ansnote === $test->ansnote)) {
            $passed = false;
            $anomalynote[] = '[NOTE expected: ' . $test->ansnote . ']';
        }

        $anomalynote = implode(' | ', $anomalynote);
        return array($passed, $errors, $rawmark, $feedback, $ansnote, $anomalynote, $trace);
    }
}
